/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <c4scripter/c4s-iconmanager.h>

#include <stdexcept>
#include "FunctionHolder.h"

#include <gtk/gtkcombobox.h>
#include <gtk/gtkcelllayout.h>

namespace
{
  const unsigned int ICON_SIZE = 16; // 16x16

  Glib::RefPtr<const Gdk::Pixbuf> get_access_icon(Gtk::Widget* widget, C4sCustomFunctionAccess access)
  {
    const gchar* stock_id;

    switch(access)
    {
    case C4S_CUSTOM_FUNCTION_ACCESS_GLOBAL:
      stock_id = C4S_STOCK_FUNC_GLOBAL;
      break;
    case C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC:
      stock_id = C4S_STOCK_FUNC_PUBLIC;
      break;
    case C4S_CUSTOM_FUNCTION_ACCESS_PROTECTED:
      stock_id = C4S_STOCK_FUNC_PROTECTED;
      break;
    case C4S_CUSTOM_FUNCTION_ACCESS_PRIVATE:
      stock_id = C4S_STOCK_FUNC_PRIVATE;
      break;
    default:
      throw std::logic_error(
        "FunctionHolder.cpp:get_access_icon:\n"
        "Invalid access was passed"
      );
    }

    GdkPixbuf* pixbuf;

    pixbuf = gtk_widget_render_icon(
      widget->gobj(),
      stock_id,
      GTK_ICON_SIZE_MENU,
      NULL
    );
    
    return Glib::wrap(pixbuf, FALSE);
  }
}

FunctionHolder::CustomColumns::CustomColumns()
{
  add(name);
  add(access);
  add(func);
}

FunctionHolder::FunctionHolder(C4sCustomFunctionList* custom_funcs, C4sEngineFunctionList* engine_funcs, const ParamInfo& param_info):
  m_custom_list(custom_funcs),
  m_engine_list(engine_funcs),
  m_param_info(param_info),
  m_custom_funcs(Gtk::ListStore::create(m_custom_columns)),
  m_custom_sort(Gtk::TreeModelSort::create(m_custom_funcs)),
  m_custom_box(m_custom_sort)
  //m_engine_box(m_engine_list.get_list())
{
  g_object_ref(G_OBJECT(m_engine_list));
  gtk_combo_box_set_model(GTK_COMBO_BOX(m_engine_box.gobj()), GTK_TREE_MODEL(c4s_engine_function_list_get_list_store(m_engine_list)));

  m_custom_sort->set_sort_column(m_custom_columns.name, Gtk::SORT_ASCENDING);
  m_custom_box.pack_start(m_custom_renderer_access);
  m_custom_box.pack_start(m_custom_renderer_name);

  m_custom_box.add_attribute(m_custom_renderer_access.property_pixbuf(), m_custom_columns.access);
  m_custom_box.add_attribute(m_custom_renderer_name.property_text(), m_custom_columns.name);

  m_custom_renderer_access.property_ypad() = 0;
  m_custom_renderer_name.property_ypad() = 0;
  g_object_set(G_OBJECT(m_custom_renderer_name.gobj()), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
  g_object_set(G_OBJECT(m_custom_renderer_name.gobj()), "width-chars", 35, NULL);

  m_engine_box.pack_start(m_engine_renderer);

  m_engine_renderer.property_ypad() = 0;
  gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(m_engine_box.gobj()), GTK_CELL_RENDERER(m_engine_renderer.gobj()), "text", 0);
  //m_engine_box.add_attribute(m_engine_renderer.property_text(), m_engine_list.get_columns().name);

  g_object_set(G_OBJECT(m_engine_renderer.gobj()), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
  g_object_set(G_OBJECT(m_engine_renderer.gobj()), "width-chars", 35, NULL);

  // When one switches between tabs and then exits the application the
  // destructor call of our base class (~HBox) leads to a segfault caused
  // by some refcounting strangeness (gtkmm-2.8.3, GTK+ 2.8.18). I could
  // however not reduce this to a simple testcase. These calls prevent that
  // error but maybe cause the combo boxes to never be freed. Still better
  // than an access violation, though. armin.
  g_object_ref(G_OBJECT(m_custom_box.gobj()));
  g_object_ref(G_OBJECT(m_engine_box.gobj()));

  pack_start(m_custom_box, Gtk::PACK_SHRINK);
  //pack_start(m_engine_box, Gtk::PACK_SHRINK);

  set_spacing(10);

  g_signal_connect_after(
    G_OBJECT(m_custom_list),
    "function-added",
    G_CALLBACK(on_function_added_static),
    this
  );
  
  g_signal_connect_after(
    G_OBJECT(m_custom_list),
    "function-removed",
    G_CALLBACK(on_function_removed_static),
    this
  );

  g_signal_connect_after(
    G_OBJECT(m_engine_list),
    "reloaded",
    G_CALLBACK(on_engine_functions_reloaded_static),
    this
  );
  //m_engine_list.signal_reloaded().connect(sigc::mem_fun(*this, &FunctionHolder::on_engine_functions_reloaded));

  m_param_info.signal_function_changed().connect(sigc::mem_fun(*this, &FunctionHolder::on_param_info_changed));
  m_param_info.signal_function_removed().connect(sigc::mem_fun(*this, &FunctionHolder::on_param_info_removed));

  m_conn_custom_changed = m_custom_box.signal_changed().connect(sigc::mem_fun(*this, &FunctionHolder::on_custom_function_changed));

  // Add initial functions
  c4s_custom_function_list_foreach(
    m_custom_list,
    on_function_added_static2,
    this
  );
//  const FunctionList::List& list = m_function_list.get_list();
//  for(FunctionList::List::const_iterator iter = list.begin(); iter != list.end(); ++ iter)
//    on_function_added(*iter);
}

FunctionHolder::~FunctionHolder()
{
  g_object_unref(G_OBJECT(m_engine_list));
}

void FunctionHolder::on_engine_functions_reloaded()
{
  gtk_combo_box_set_model(GTK_COMBO_BOX(m_engine_box.gobj()), GTK_TREE_MODEL(c4s_engine_function_list_get_list_store(m_engine_list)));
//  m_engine_box.set_model(m_engine_list.get_list());

  // TODO: Get current function from ParamInfo, update.
}

void FunctionHolder::on_function_added(const C4sCustomFunction* func)
{
  Gtk::TreeIter iter = m_custom_funcs->append();
  (*iter)[m_custom_columns.name] = c4s_function_get_name(C4S_FUNCTION(func));
  (*iter)[m_custom_columns.access] = get_access_icon(this, c4s_custom_function_get_access(func));
  (*iter)[m_custom_columns.func] = func;

  g_signal_connect_after(
    G_OBJECT(func),
    "name-changed",
    G_CALLBACK(on_function_name_changed_static),
    this
  );

  g_signal_connect_after(
    G_OBJECT(func),
    "access-changed",
    G_CALLBACK(on_function_access_changed_static),
    this
  );
}

void FunctionHolder::on_function_removed(const C4sCustomFunction* func)
{
        Gtk::TreeIter child = find_function(func);
        m_custom_funcs->erase(child);
}

void FunctionHolder::on_function_access_changed_static(const C4sCustomFunction* func, C4sCustomFunctionAccess access, gpointer user_data)
{
  static_cast<FunctionHolder*>(user_data)->on_function_access_changed(func);
}

void FunctionHolder::on_function_name_changed_static(const C4sCustomFunction* func, const gchar* name, gpointer user_data)
{
  static_cast<FunctionHolder*>(user_data)->on_function_name_changed(func);
}

void FunctionHolder::on_function_access_changed(const C4sCustomFunction* func)
{
  Gtk::TreeIter child = find_function(func);
  (*child)[m_custom_columns.access] = get_access_icon(this, c4s_custom_function_get_access(func));
}

void FunctionHolder::on_function_name_changed(const C4sCustomFunction* func)
{
  Gtk::TreeIter child = find_function(func);
  (*child)[m_custom_columns.name] = c4s_function_get_name(C4S_FUNCTION(func));
}

void FunctionHolder::on_param_info_changed(const C4sFunction* func, unsigned int param)
{
  Gtk::TreeNodeChildren children = m_custom_sort->children();
  Gtk::TreeIter iter;

  m_conn_custom_changed.block();
  for(iter = children.begin(); iter != children.end(); ++ iter)
  {
    const C4sCustomFunction* custom_func = (*iter)[m_custom_columns.func];
    if(C4S_FUNCTION(custom_func) == func)
    {
      m_custom_box.set_active(iter);
      break;
    }
  }

  if(iter == children.end())
    m_custom_box.set_active(-1);

  m_conn_custom_changed.unblock();
/*
  Glib::RefPtr<Gtk::ListStore> engine_funcs = m_engine_list.get_list();
  children = engine_funcs->children();
  for(Gtk::TreeIter iter = children.begin(); iter != children.end(); ++ iter)
  {
    if( (*iter)[m_engine_list.get_columns().func] == &func)
    {
      m_custom_box.set_active(-1);
      m_engine_box.set_active(iter);
      return;
    }
  }
*/
}

void FunctionHolder::on_param_info_removed()
{
  m_conn_custom_changed.block();
  m_custom_box.set_active(-1);
  m_conn_custom_changed.unblock();
  //m_engine_box.set_active(-1);
}

void FunctionHolder::on_custom_function_changed()
{
  Gtk::TreeIter iter = m_custom_box.get_active();
  m_conn_custom_changed.block();
  m_custom_box.set_active(-1);
  m_conn_custom_changed.unblock();

  const C4sCustomFunction* func = (*iter)[m_custom_columns.func];
  m_signal_function_selected.emit(func);
}

Gtk::TreeIter FunctionHolder::find_function(const C4sCustomFunction* func)
{
  Gtk::TreeNodeChildren children = m_custom_funcs->children();

  for(Gtk::TreeIter iter = children.begin(); iter != children.end(); ++ iter)
  {
    if( (*iter)[m_custom_columns.func] == func)
      return iter;
  }

  throw std::logic_error(std::string("Function ") + c4s_function_get_name(C4S_FUNCTION(func)) + " not found in custom function list");
}
