/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "MainWindow.h"

#include <c4scripter/storage/c4s-storage.h>

#ifdef _WIN32
# include <windows.h>
#endif

MainWindow::MainWindow(int argc, char* argv[], C4sRemoteComm* comm) :
  Gtk::Window(Gtk::WINDOW_TOPLEVEL),
  m_icon_manager(c4s_icon_manager_new()),
  m_preferences(c4s_preferences_new(&C4S_STORAGE_KEYFILE)),
  m_file_holder(m_preferences, *this),
  m_file_handler(m_main_menu, m_file_holder, *this),
  m_doc_control(m_main_menu, m_file_holder, m_preferences, *this),
  m_view_control(m_main_menu, m_preferences, *this),
  m_statusbar(m_file_holder, m_preferences),
  m_paramtip(m_file_holder, m_preferences, *this)
{
  add_events(Gdk::FOCUS_CHANGE_MASK | Gdk::STRUCTURE_MASK);

  m_main_menu.action_quit->signal_activate().connect(sigc::mem_fun(*this, &MainWindow::on_quit));
  m_main_menu.action_help_about->signal_activate().connect(sigc::mem_fun(*this, &MainWindow::on_about));
  m_main_menu.action_help_documentation->signal_activate().connect(sigc::mem_fun(*this, &MainWindow::on_documentation));
  
  m_logo = gtk_widget_render_icon(
    GTK_WIDGET(gobj()),
    C4S_STOCK_LOGO,
    (GtkIconSize)-1,
    NULL
  );

  std::deque<Glib::RefPtr<Gdk::Pixbuf> > icons;
  icons.push_back(Glib::wrap(m_logo, FALSE));
  set_default_icon_list(icons);

  m_file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &MainWindow::on_document_changed));

  set_default_size(800, 600);
  add_accel_group(m_main_menu.get_accel_group());

  m_top_box.pack_start(m_main_menu, Gtk::PACK_SHRINK);
  m_top_box.pack_start(m_file_holder, Gtk::PACK_EXPAND_WIDGET);
  m_top_box.pack_start(m_statusbar, Gtk::PACK_SHRINK);

  for(int i = 0; i < argc; ++ i)
    m_file_handler.open_document_from(argv[i]);

  if(argc == 0)
    m_file_handler.new_document();

  g_signal_connect(
    G_OBJECT(comm),
    "file-received",
    G_CALLBACK(on_file_received_static),
    this
  );

  add(m_top_box);
  show_all();
}

MainWindow::~MainWindow()
{
  g_object_unref(m_icon_manager);
  g_object_unref(m_logo);

  c4s_preferences_save(m_preferences);
  g_object_unref(m_preferences);
}

bool MainWindow::on_delete_event(GdkEventAny* event)
{
  return !m_file_handler.close_all_documents();
}

void MainWindow::on_quit()
{
  if(m_file_handler.close_all_documents())
    hide();
}

void MainWindow::on_about()
{
  static const gchar* artists[] = { "Matthias Rottländer <MatthiasRo@web.de>", "Michael Hohenstein <Mikiman@icqmail.com>", NULL };
  static const gchar* authors[] = { "Bernhard Bonigl <boni@bbgen.net>", "Armin Burgmeier <armin@arbur.net>", NULL };
  static const gchar* comments = "C4Script editor";
  static const gchar* copyright = "Copyright © 2006 Bernhard Bonigl, Armin Burgmeier";
  static const GdkPixbuf* logo = m_logo;
  static const gchar* name = "C4Scripter";
  static const gchar* version = PACKAGE_VERSION;

  gtk_show_about_dialog(gobj(), "artists", artists, "authors", authors, "comments", comments, "copyright", copyright, "name", name, "logo", logo, "version", version, NULL);
}

void MainWindow::on_file_received(const std::string& file)
{
  m_file_handler.open_document_from(file);
  present(); // Bring to foreground
}

void MainWindow::on_document_changed(DocumentHolder* holder)
{
  if(m_title_conn.connected()) m_title_conn.disconnect();
  if(m_modified_conn.connected()) m_modified_conn.disconnect();

  if(holder != NULL)
  {
    m_modified_conn = holder->get_document().get_text_buffer()->signal_modified_changed().connect(sigc::mem_fun(*this, &MainWindow::on_modified_changed));
    m_title_conn = holder->get_document().signal_title_changed().connect(sigc::mem_fun(*this, &MainWindow::on_title_changed));
  }

  update_title();
}

void MainWindow::on_modified_changed()
{
  update_title();
}

void MainWindow::on_title_changed()
{
  update_title();
}

void MainWindow::update_title()
{
  DocumentHolder* holder = m_file_holder.get_current_document();
  if(holder == NULL)
  {
    set_title("C4Scripter");
  }
  else
  {
    std::string title = holder->get_document().get_title();
    if(holder->get_document().get_text_buffer()->get_modified())
      title = "*" + title;

    set_title(title + " - C4Scripter");
  }
}

void MainWindow::on_documentation()
{
  DocumentHolder* holder = m_file_holder.get_current_document();
  if(holder != NULL)
  {
    gchar* path;
    const gchar* chm_path;

    const C4sFunction* function = holder->get_document().get_param_info().get_current_function();
    
    chm_path = c4s_preferences_get_chm_file_path(m_preferences);
    if(chm_path != NULL && *chm_path != '\0')
    {
#ifdef _WIN32
      char win_path[MAX_PATH + 1];
      GetWindowsDirectory(win_path, MAX_PATH);

      if(function == NULL)
      {
        path = g_strdup_printf("%s\\hh.exe %s", win_path, chm_path);
      }
      else
      {
        path = g_strdup_printf("%s\\hh.exe %s::/sdk/script/fn/%s.html", win_path, chm_path, c4s_function_get_name(function));
      }
      
      WinExec(path, SW_SHOW);
      g_free(path);
#endif
    }
    else
    {
#ifdef _WIN32
      if(function == NULL)
      {
        path = g_strdup("http://www.clonk.de/docs/en/sdk/index.html");
      }
      else
      {
        path = g_strdup_printf("http://www.clonk.de/docs/en/sdk/script/fn/%s.html", c4s_function_get_name(function));
      }
      
      ShellExecute(NULL, "open", path, NULL, NULL, SW_SHOW);
      g_free(path);
#endif
    }
  }
}
