/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-textutil.h>

#include "ParamInfo.h"

ParamInfo::ParamInfo(const Glib::RefPtr<Gtk::TextBuffer>& buffer, const C4sHighlighter* highlighter, C4sCustomFunctionList* custom_function_list, C4sEngineFunctionList* engine_function_list):
  m_buffer(buffer),
  m_highlighter(highlighter),
  m_custom_function_list(custom_function_list),
  m_engine_function_list(engine_function_list),
  cur_function(NULL), cur_param(0)
{
  g_object_ref(G_OBJECT(highlighter));

  g_signal_connect_after(
    G_OBJECT(engine_function_list),
    "reloaded",
    G_CALLBACK(on_engine_functions_reloaded_static),
    this
  );

  g_object_ref(G_OBJECT(m_custom_function_list));
  g_object_ref(G_OBJECT(m_engine_function_list));
  //engine_function_list.signal_reloaded().connect(sigc::mem_fun(*this, &ParamInfo::on_engine_functions_reloaded));
  m_buffer->signal_mark_set().connect(sigc::mem_fun(*this, &ParamInfo::on_cursor_moved));

  g_signal_connect_after(
    G_OBJECT(highlighter),
    "highlighted",
    G_CALLBACK(on_highlighted_static),
    this
  );
//  highlighter.signal_highlighted().connect(sigc::mem_fun(*this, &ParamInfo::on_highlighted));
}

ParamInfo::~ParamInfo()
{
  g_object_unref(G_OBJECT(m_highlighter));
  g_object_unref(G_OBJECT(m_custom_function_list));
  g_object_unref(G_OBJECT(m_engine_function_list));
}

void ParamInfo::on_engine_functions_reloaded()
{
  // We must not reference the current function anymore, so we no longer know
  // the function name of the current function, full rescan.
  update();
}

void ParamInfo::on_cursor_moved(const Gtk::TextIter& iter, const Glib::RefPtr<Gtk::TextMark>& mark)
{
  if(mark == m_buffer->get_insert())
  {
    update();
  }
}

void ParamInfo::on_highlighted_static(C4sHighlighter* highlighter, const GtkTextIter* from, const GtkTextIter* to, gpointer user_data)
{
  static_cast<ParamInfo*>(user_data)->on_highlighted(Glib::wrap(from), Glib::wrap(to));
}

void ParamInfo::on_highlighted(const Gtk::TextIter& from, const Gtk::TextIter& to)
{
  // Document has been rehighlighted, so document has changed -> update current function
  update();
}

void ParamInfo::update()
{
  Gtk::TextIter pos = m_buffer->get_insert()->get_iter();

  // Move out of comment or string
  if(c4s_highlighter_is_comment(m_highlighter, pos.gobj()))
    c4s_highlighter_backward_out_of_comment(m_highlighter, pos.gobj());
  else if(c4s_highlighter_is_string(m_highlighter, pos.gobj()))
    c4s_highlighter_backward_out_of_string(m_highlighter, pos.gobj());

  int level = 0;
  unsigned int param = 0;

  while(!pos.is_start())
  {
    if(c4s_highlighter_ends_comment(m_highlighter, pos.gobj()))
      c4s_highlighter_backward_out_of_comment(m_highlighter, pos.gobj());
    else if(c4s_highlighter_ends_string(m_highlighter, pos.gobj()))
    {
      if(level == -1) break;
      c4s_highlighter_backward_out_of_string(m_highlighter, pos.gobj());
    }
    else
    {
      -- pos;
      gunichar c = *pos;

      // Ignore spaces
      if(Glib::Unicode::isspace(c))
        continue;

      if(level == -1)
      {
        if(Glib::Unicode::isalnum(c) || c == '_')
        {
          Gtk::TextIter end = pos; ++ end;
          c4s_backward_identifier(pos.gobj());
          
          std::string func = pos.get_slice(end);
          const C4sFunction* function;
          if(!(function = C4S_FUNCTION(c4s_custom_function_list_find_function(m_custom_function_list, func.c_str()))))
            if(!(function = C4S_FUNCTION(c4s_engine_function_list_find_function(m_engine_function_list, func.c_str()))))
              break;

          cur_function = function;
          cur_param = param;

          m_signal_function_changed.emit(cur_function, cur_param);
	  return;
        }

        // Was not a function paranthesis
        level = 0;
        param = 0;
      }

      if(c == ')')
        ++ level;
      else if(c == '(')
        -- level;
      else if(c == ',' && level == 0)
        ++ param;
      else if(c == ';' || c == '{' || c == '}' || c == ':')
        break;
    }
  }

  // No valid function found
  if(cur_function == NULL) return;

  cur_function = NULL;
  m_signal_function_removed.emit();
}
