/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ParamView.h"
#include "ParamInfo.h"

struct C4sParamViewPrivate_ {
	ParamInfo* param_info;
	C4sPreferences* preferences;
	GdkRectangle tip_rect;
	PangoLayout* layout;
};

static GObjectClass* parent_class = NULL;

static void c4s_param_view_class_init(C4sParamViewClass* klass);
static void c4s_param_view_init(C4sParamView* view);
static void c4s_param_view_finalize(GObject* object);
static gint c4s_param_view_expose(GtkWidget* widget, GdkEventExpose* event);

static void
c4s_param_view_class_init(C4sParamViewClass* klass)
{
	GObjectClass* object_class;
	GtkWidgetClass* widget_class;

	object_class = G_OBJECT_CLASS(klass);
	parent_class = (GObjectClass*)g_type_class_peek_parent(klass);
	widget_class = GTK_WIDGET_CLASS(klass);

	widget_class->expose_event = c4s_param_view_expose;
	object_class->finalize = c4s_param_view_finalize;
}

static void
c4s_param_view_init(C4sParamView* view)
{
	view->priv = g_new0(C4sParamViewPrivate, 1);
	view->priv->tip_rect.x = 0;
	view->priv->tip_rect.y = 0;
	view->priv->tip_rect.width = 0;
	view->priv->tip_rect.height = 0;
}

static void
c4s_param_view_finalize(GObject* object)
{
	C4sParamView* view;
	view = C4S_PARAM_VIEW(object);

	if(view->priv->layout != NULL)
		g_object_unref(view->priv->layout);

	g_free(view->priv);
	G_OBJECT_CLASS(parent_class)->finalize(object);
}

static gint c4s_param_view_expose(GtkWidget* widget, GdkEventExpose* event)
{
	C4sParamView* view;
	GdkRectangle redraw_rect;
	GdkRectangle intersection;
	gboolean result;
	gboolean event_handled;
	GdkWindow* window;
	cairo_t* cr;

	view = C4S_PARAM_VIEW(widget);

	if(GTK_WIDGET_CLASS(parent_class)->expose_event)
	{
		event_handled = GTK_WIDGET_CLASS(parent_class)->expose_event(
			widget,
			event
		);
	}

	window = gtk_text_view_get_window(
		GTK_TEXT_VIEW(view),
		GTK_TEXT_WINDOW_TEXT
	);

	if(event->window != window) return event_handled;
	if(!c4s_preferences_get_view_tooltips(view->priv->preferences))
		return event_handled;

	gtk_text_view_buffer_to_window_coords(
		GTK_TEXT_VIEW(view),
		GTK_TEXT_WINDOW_TEXT,
		view->priv->tip_rect.x,
		view->priv->tip_rect.y,
		&redraw_rect.x,
		&redraw_rect.y
	);

	redraw_rect.width = view->priv->tip_rect.width;
	redraw_rect.height = view->priv->tip_rect.height;

	result = gdk_rectangle_intersect(
		&event->area,
		&redraw_rect,
		&intersection
	);

	if(result == FALSE) return event_handled;

	cr = gdk_cairo_create(window);

	cairo_rectangle(
		cr,
		intersection.x,
		intersection.y,
		intersection.width,
		intersection.height
	);

	cairo_clip(cr);
	cairo_set_line_width(cr, 1.0);

	cairo_rectangle(
		cr,
		redraw_rect.x,
		redraw_rect.y,
		redraw_rect.width,
		redraw_rect.height
	);

	cairo_set_source_rgb(cr, 1.0, 1.0, 192.0/255.0);
	cairo_fill_preserve(cr);
	cairo_set_source_rgb(cr, 0, 0, 0);
	cairo_stroke_preserve(cr);
	cairo_destroy(cr);

	/* TODO: Could we also use cairo to draw this? */
	gtk_paint_layout(
		widget->style,
		window,
		GtkStateType(GTK_WIDGET_STATE(widget)),
		FALSE,
		&intersection,
		widget,
		NULL,
		redraw_rect.x + 2,
		redraw_rect.y + 2,
		view->priv->layout
	);

	return event_handled;
}

static gchar* c4s_param_view_create_tip_text(const C4sFunction* func, unsigned int param, gsize* param_len, gsize* full_len)
{
	GString* str;
	unsigned int i;
	C4sValueType type;
	const C4sFunctionParameter* parameter;
	gsize n_parameter;

	str = g_string_sized_new(128);

	type = c4s_function_get_return_type(func);

	g_string_append(
	  str,
	  c4s_value_type_to_string(C4sValueType((int)type & ~(int)C4S_VALUE_TYPE_REF))
	);

	if((int)type & (int)C4S_VALUE_TYPE_REF)
	  g_string_append(str, "&amp;");

	g_string_append_c(str, ' ');
	g_string_append(str, c4s_function_get_name(func));
	g_string_append_c(str, '(');

	parameter = c4s_function_get_parameters(func, &n_parameter);

	for(i = 0; i < n_parameter; ++ i)
	{
		if(i > 0) g_string_append(str, ", ");

		if(i == param)
		{
			*param_len = str->len;
			g_string_append(str, "<b>");
		}

		g_string_append(
			str,
			c4s_value_type_to_string(C4sValueType((int)parameter[i].type & ~(int)C4S_VALUE_TYPE_REF))
		);

		if((int)parameter[i].type & (int)C4S_VALUE_TYPE_REF)
			g_string_append(str, "&amp;");

		g_string_append_c(str, ' ');
		g_string_append(str, parameter[i].name);

		if(i == param) g_string_append(str, "</b>");
	}

	if(param >= n_parameter)
		*param_len = str->len;

	g_string_append_c(str, ')');
	*full_len = str->len;
	return g_string_free(str, FALSE);
}

static void c4s_param_view_redraw_tip_rect(C4sParamView* view)
{
	GdkRectangle redraw_rect;
	GdkWindow* window;

	/* Nothing to draw */
	if(view->priv->tip_rect.width == 0)
		return;

	gtk_text_view_buffer_to_window_coords(
		GTK_TEXT_VIEW(view),
		GTK_TEXT_WINDOW_TEXT,
		view->priv->tip_rect.x,
		view->priv->tip_rect.y,
		&redraw_rect.x,
		&redraw_rect.y
	);

	redraw_rect.width = view->priv->tip_rect.width;
	redraw_rect.height = view->priv->tip_rect.height;

	window = gtk_text_view_get_window(
		GTK_TEXT_VIEW(view),
		GTK_TEXT_WINDOW_TEXT
	);

	gdk_window_invalidate_rect(
		window,
		&redraw_rect,
		FALSE
	);
}

static void c4s_param_view_recalc_tip_rect(C4sParamView* view)
{
	gchar* text;
	gsize param_len, full_len;
	int offset_width, text_width, text_height;
	GtkTextIter insert_iter;
	GtkTextBuffer* buffer;
	GdkRectangle iter_location;

	/* No function selected */
	if(view->priv->param_info->get_current_function() == NULL)
	{
		view->priv->tip_rect.x = 0;
		view->priv->tip_rect.y = 0;
		view->priv->tip_rect.width = 0;
		view->priv->tip_rect.height = 0;
		return;
	}

	const C4sFunction* func = view->priv->param_info->get_current_function();
	unsigned int param = view->priv->param_info->get_current_parameter();

	text = c4s_param_view_create_tip_text(
		func,
		param,
		&param_len,
		&full_len
	);

	if(view->priv->layout == NULL)
	{
		view->priv->layout = gtk_widget_create_pango_layout(
			GTK_WIDGET(view),
			NULL
		);
	}

	pango_layout_set_markup(view->priv->layout, text, param_len);

	pango_layout_get_pixel_size(
		view->priv->layout,
		&offset_width,
		&text_height
	);

	/* TODO: Would be enough to query remaining extent? */
	pango_layout_set_markup(view->priv->layout, text, full_len);
	pango_layout_get_pixel_size(view->priv->layout, &text_width, NULL);
	g_free(text);

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

	gtk_text_buffer_get_iter_at_mark(
		buffer,
		&insert_iter,
		gtk_text_buffer_get_insert(buffer)
	);

	gtk_text_view_get_iter_location(
		GTK_TEXT_VIEW(view),
		&insert_iter,
		&iter_location
	);

	view->priv->tip_rect.x = iter_location.x - (offset_width + 2);
	view->priv->tip_rect.y = iter_location.y + iter_location.height;
	view->priv->tip_rect.width = text_width + 4;
	view->priv->tip_rect.height = text_height + 4;
}

static void c4s_param_view_function_changed_cb(const C4sFunction* func, unsigned int param, C4sParamView* view)
{
	if(c4s_preferences_get_view_tooltips(view->priv->preferences))
		c4s_param_view_redraw_tip_rect(view);

	c4s_param_view_recalc_tip_rect(view);

	if(c4s_preferences_get_view_tooltips(view->priv->preferences))
		c4s_param_view_redraw_tip_rect(view);
}

static void c4s_param_view_notify_cb(GObject* object, GParamSpec* spec, gpointer data)
{
	if(strcmp(spec->name, "view-tooltips") == 0)
		c4s_param_view_redraw_tip_rect((C4sParamView*)data);
}

static void c4s_param_view_style_changed_cb(C4sParamView* view, GtkStyle* style)
{
	if(c4s_preferences_get_view_tooltips(view->priv->preferences))
		c4s_param_view_redraw_tip_rect(view);
	
	c4s_param_view_recalc_tip_rect(view);

	if(c4s_preferences_get_view_tooltips(view->priv->preferences))
		c4s_param_view_redraw_tip_rect(view);
}

static void c4s_param_view_function_removed_cb(C4sParamView* view)
{
	c4s_param_view_redraw_tip_rect(view); /* clear */
	c4s_param_view_recalc_tip_rect(view); /* reset */
}

GType c4s_param_view_get_type(void)
{
	static GType our_type = 0;

	if(our_type == 0)
	{
		static const GTypeInfo our_info = {
			sizeof(C4sParamViewClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) c4s_param_view_class_init,
			NULL,
			NULL,
			sizeof(C4sParamView),
			0,
			(GInstanceInitFunc) c4s_param_view_init
		};

		our_type = g_type_register_static(
			GTK_TYPE_SOURCE_VIEW,
			"C4sParamView",
			&our_info,
			GTypeFlags(0)
		);
	}

	return our_type;
}

GtkWidget* c4s_param_view_new(void* param_info, C4sPreferences* preferences)
{
	GtkWidget* widget;
	GtkSourceBuffer* buffer;

	buffer = gtk_source_buffer_new(NULL);
	widget = c4s_param_view_new_with_buffer(
		buffer,
		param_info,
		preferences
	);

	return widget;
}

GtkWidget* c4s_param_view_new_with_buffer(GtkSourceBuffer* buffer,
                                          void* param_info,
                                          C4sPreferences* preferences)
{
	GtkWidget* view;
	ParamInfo* info;

	view = (GtkWidget*)g_object_new(C4S_TYPE_PARAM_VIEW, NULL);

	gtk_text_view_set_buffer(
		GTK_TEXT_VIEW(view),
		GTK_TEXT_BUFFER(buffer)
	);

	g_object_unref(buffer);

	info = (ParamInfo*)param_info;

	C4S_PARAM_VIEW(view)->priv->param_info = info;
	C4S_PARAM_VIEW(view)->priv->preferences = preferences;

	info->signal_function_changed().connect(sigc::bind(sigc::ptr_fun(c4s_param_view_function_changed_cb), C4S_PARAM_VIEW(view)));
	info->signal_function_removed().connect(sigc::bind(sigc::ptr_fun(c4s_param_view_function_removed_cb), C4S_PARAM_VIEW(view)));

	g_signal_connect(
		G_OBJECT(preferences),
		"notify",
		G_CALLBACK(c4s_param_view_notify_cb),
		view
	);

	g_signal_connect(G_OBJECT(view), "style-set", G_CALLBACK(c4s_param_view_style_changed_cb), NULL);

	return view;
}
