/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-customfunctionlist.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-marshal.h>

#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextmark.h>

#include <string.h>

typedef struct _C4sCustomFunctionListPrivate C4sCustomFunctionListPrivate;
struct _C4sCustomFunctionListPrivate
{
  GtkTextBuffer* buffer;
  C4sHighlighter* highlighter;

  GList* function_list;
  GHashTable* function_table; /* lookup by name */
  
  GList* partial_functions;
};

#define C4S_CUSTOM_FUNCTION_LIST_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_CUSTOM_FUNCTION_LIST, \
    C4sCustomFunctionListPrivate \
  ))

enum {
  PROP_0,

  PROP_BUFFER,
  PROP_HIGHLIGHTER
};

enum {
  FUNCTION_ADDED,
  FUNCTION_REMOVED,
  LAST_SIGNAL
};

static GObjectClass* parent_class = NULL;
static guint custom_function_list_signals[LAST_SIGNAL];

static C4sCustomFunction*
c4s_custom_function_list_find_custom_function(C4sCustomFunctionList* list,
                                              const GtkTextIter* pos,
                                              GList* limit)
{
  C4sCustomFunctionListPrivate* priv;
  GtkTextIter begin;
  GList* item;

  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);
  for(item = priv->function_list; item != limit; item = item->next)
  {
    c4s_custom_function_get_begin(C4S_CUSTOM_FUNCTION(item->data), &begin);
    if(gtk_text_iter_compare(pos, &begin) == 0)
      return C4S_CUSTOM_FUNCTION(item->data);
  }
  
  return NULL;
}

static GtkTextMark*
c4s_custom_function_list_find_partial_function(C4sCustomFunctionList* list,
                                               const GtkTextIter* pos,
                                               GList* limit)
{
  C4sCustomFunctionListPrivate* priv;
  GtkTextIter func_pos;
  GList* item;
  
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);
  for(item = priv->partial_functions; item != limit; item = item->next)
  {
    gtk_text_buffer_get_iter_at_mark(
      priv->buffer,
      &func_pos,
      GTK_TEXT_MARK(item->data)
    );
    
    if(gtk_text_iter_compare(&func_pos, pos) == 0)
      return GTK_TEXT_MARK(item->data);
  }
  
  return NULL;
}

static void
c4s_custom_function_list_update_partial_functions(C4sCustomFunctionList* list,
                                                  G_GNUC_UNUSED
                                                  const GtkTextIter* begin,
                                                  const GtkTextIter* end)
{
  C4sCustomFunctionListPrivate* priv;
  C4sCustomFunction* function;
  GtkTextIter func_begin;
  GtkTextIter func_end;
  GList* item;
  GList* next;
  gchar* func_str;
  
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);
  for(item = priv->partial_functions; item != NULL; item = next)
  {
    next = item->next;

    gtk_text_buffer_get_iter_at_mark(
      priv->buffer,
      &func_begin,
      GTK_TEXT_MARK(item->data)
    );

    /* func may be affected if change was behind; however, this is
     * still suboptimal, we should limit range */
    if(gtk_text_iter_compare(end, &func_begin) >= 0)
    {
      func_end = func_begin;
      c4s_forward_identifier(&func_end);

      func_str = gtk_text_buffer_get_slice(
        priv->buffer,
        &func_begin,
        &func_end,
        TRUE
      );

      /* TODO: Shouldn't we check for duplicates here? If the text is
       * "func func foo", and one deletes the first func, we have two
       * partial functions, haven't we? */

      if(c4s_starts_identifier(&func_begin) == FALSE ||
         strcmp(func_str, "func") != 0 ||
         c4s_highlighter_is_code(priv->highlighter, &func_begin) == FALSE)
      {
        /* No partial function anymore */
        gtk_text_buffer_delete_mark(priv->buffer, GTK_TEXT_MARK(item->data));

        priv->partial_functions = g_list_delete_link(
          priv->partial_functions,
          item
        );
      }
      else
      {
        function = c4s_custom_function_new(&func_begin, priv->highlighter);
        if(function != NULL)
        {
          /* Partial function has been completed */
          gtk_text_buffer_delete_mark(
            priv->buffer,
            GTK_TEXT_MARK(item->data)
          );

          priv->partial_functions = g_list_delete_link(
            priv->partial_functions,
            item
          );

          g_signal_emit(
            G_OBJECT(list),
            custom_function_list_signals[FUNCTION_ADDED],
            0,
            function
          );

          /* Default handler should have ref'ed function */
          g_object_unref(G_OBJECT(function));
        }
      }
      
      g_free(func_str);
    }
  }
}

static void
c4s_custom_function_list_update_existing_functions(C4sCustomFunctionList* list,
                                                   const GtkTextIter* begin,
                                                   const GtkTextIter* end)
{
  C4sCustomFunctionListPrivate* priv;
  C4sCustomFunction* function;
  C4sCustomFunction* other_function;
  GtkTextMark* partial_function;
  GtkTextIter func_access_begin;
  GtkTextIter func_begin;
  GtkTextIter func_end;
  GList* item;
  GList* next;
  gchar* func_str;
  
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  /* It is quite inefficient to have to traverse the whole list here. We
   * could use a GPtrArray instead, sort it by apperance in the script and
   * use binary search here. */
  for(item = priv->function_list; item != NULL; item = next)
  {
    next = item->next;

    function = C4S_CUSTOM_FUNCTION(item->data);    
    c4s_custom_function_get_end(function, &func_end);
    c4s_custom_function_get_access_begin(function, &func_access_begin);

    /* must be <= respectively >= in both cases because all marks could
     * have been moved to the same position when a complete function has
     * been erased at once. */
    if(gtk_text_iter_compare(begin, &func_end) <= 0 &&
       gtk_text_iter_compare(end, &func_access_begin) >= 0)
    {
      c4s_custom_function_get_begin(function, &func_begin);

      /* Another function may have moved to the same location as this one if
       * much text with multiple functions in it has been deleted at once, so
       * if we handled a function before this one at the same location, we no
       * longer need this function. */
      other_function = c4s_custom_function_list_find_custom_function(
        list,
        &func_begin,
        item
      );

      /* Found such a function, so this one can be removed. It also may
       * be removed if it is not a function anymore (update returns FALSE) */
      if(other_function != NULL ||
         c4s_custom_function_update(function, priv->highlighter) == FALSE)
      {
        g_signal_emit(
          G_OBJECT(list),
          custom_function_list_signals[FUNCTION_REMOVED],
          0,
          function
        );

        /* This is only true when we did non found a duplicate function, so
         * the code below is only executed when c4s_custom_function_update
         * failed. If the function has been removed because of a duplicate,
         * we don't even need a partial function here because there is
         * already a function. */
        if(other_function == NULL && c4s_starts_identifier(&func_begin))
        {
          func_end = func_begin;
          c4s_forward_identifier(&func_end);
          
          func_str = gtk_text_buffer_get_slice(
            priv->buffer,
            &func_begin,
            &func_end,
            TRUE
          );

          if(strcmp(func_str, "func") == 0 &&
             c4s_highlighter_is_code(priv->highlighter, &func_begin))
          {
            /* It is however not sure whether there is already another
             * partial function at this position */
            partial_function = c4s_custom_function_list_find_partial_function(
              list,
              &func_begin,
              NULL
            );
            
            if(partial_function == NULL)
            {
              partial_function = gtk_text_buffer_create_mark(
                priv->buffer,
                NULL,
                &func_begin,
                FALSE
              );

              priv->partial_functions = g_list_prepend(
                priv->partial_functions,
                partial_function
              );
            }
          }

          g_free(func_str);
        }
      }
    }
  }
}

static void
c4s_custom_function_list_check_new_functions(C4sCustomFunctionList* list,
                                             const GtkTextIter* begin,
                                             const GtkTextIter* end)
{
  C4sCustomFunctionListPrivate* priv;
  C4sCustomFunction* function;
  GtkTextMark* partial_function;
  GtkTextIter start;
  GtkTextIter stop;
  GtkTextIter match_start;
  GtkTextIter match_end;
  
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);
  start = *begin;
  stop = *end;

  while(c4s_forward_search(&start, "func", 0, &match_start, &match_end, &stop))
  {
    function = c4s_custom_function_list_find_custom_function(
      list,
      &start,
      NULL
    );

    if(function == NULL)
    {
      /* No custom function here, so make one */
      function = c4s_custom_function_new(&match_start, priv->highlighter);
      if(function != NULL)
      {
        /* match_start has been updated to point behind the function
         * definition, and this is were we want to continue for the
         * next iteration. */
        start = match_start;

        g_signal_emit(
          G_OBJECT(list),
          custom_function_list_signals[FUNCTION_ADDED],
          0,
          function
        );
      }
      else
      {
        /* Function creation failed, could be partial function */
        if(c4s_starts_identifier(&match_start) &&
           c4s_ends_identifier(&match_end) &&
           c4s_highlighter_is_code(priv->highlighter, &match_start))
        {
          partial_function = c4s_custom_function_list_find_partial_function(
            list,
            &match_start,
            NULL
          );

          if(partial_function == NULL)
          {
            partial_function = gtk_text_buffer_create_mark(
              priv->buffer,
              NULL,
              &match_start,
              FALSE
            );

            priv->partial_functions = g_list_prepend(
              priv->partial_functions,
              partial_function
            );
          }
        }

        /* Continue behind func keyword for partial func */
        start = match_end;
      }
    }
    else
    {
      /* Continue behind existing function */
      c4s_custom_function_get_end(function, &start);
    }
  }
}

static void
c4s_custom_function_list_update(C4sCustomFunctionList* list,
                                const GtkTextIter* begin,
                                const GtkTextIter* end)
{
  c4s_custom_function_list_update_partial_functions(list, begin, end);
  c4s_custom_function_list_update_existing_functions(list, begin, end);
  c4s_custom_function_list_check_new_functions(list, begin, end);
}

static void
c4s_custom_function_list_renew(C4sCustomFunctionList* list)
{
  C4sCustomFunctionListPrivate* priv;
  GtkTextIter start;
  GtkTextIter end;

  /* TODO: We should remove any functions and partial functions here */

  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);  
  if(priv->buffer != NULL && priv->highlighter != NULL)
  {
    gtk_text_buffer_get_start_iter(priv->buffer, &start);
    gtk_text_buffer_get_end_iter(priv->buffer, &end);
    
    c4s_custom_function_list_update(list, &start, &end);
  }
}

static void
c4s_custom_function_list_highlighted_cb(G_GNUC_UNUSED
                                        C4sHighlighter* highlighter,
                                        GtkTextIter* begin,
                                        GtkTextIter* end,
                                        gpointer user_data)
{
  C4sCustomFunctionList* list;
  list = C4S_CUSTOM_FUNCTION_LIST(user_data);

  c4s_custom_function_list_update(list, begin, end);
}

static void
c4s_custom_function_list_init(C4sCustomFunctionList* list)
{
  C4sCustomFunctionListPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  priv->buffer = NULL;
  priv->highlighter = NULL;

  priv->function_list = NULL;
  priv->function_table = g_hash_table_new(g_str_hash, g_str_equal);
  priv->partial_functions = NULL;
}

static void 
c4s_custom_function_list_finalize(GObject* object)
{
  C4sCustomFunctionList* list;
  C4sCustomFunctionListPrivate* priv;
  GList* item;

  list = C4S_CUSTOM_FUNCTION_LIST(object);
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  g_hash_table_destroy(priv->function_table);
  
  for(item = priv->function_list; item != NULL; item = item->next)
    g_object_unref(G_OBJECT(item->data));
  g_list_free(priv->function_list);
  
  for(item = priv->partial_functions; item != NULL; item = item->next)
    gtk_text_buffer_delete_mark(priv->buffer, GTK_TEXT_MARK(item->data));
  g_list_free(priv->partial_functions);

  if(priv->buffer != NULL) g_object_unref(G_OBJECT(priv->buffer));
  if(priv->highlighter != NULL) g_object_unref(G_OBJECT(priv->highlighter));

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_custom_function_list_set_property(GObject* object,
                                      guint prop_id,
                                      const GValue* value,
                                      GParamSpec* pspec)
{
  C4sCustomFunctionList* list;
  C4sCustomFunctionListPrivate* priv;

  g_return_if_fail(C4S_IS_CUSTOM_FUNCTION_LIST(object));

  list = C4S_CUSTOM_FUNCTION_LIST(object);
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  switch(prop_id)
  {
  case PROP_BUFFER:
    if(priv->buffer != NULL) g_object_unref(G_OBJECT(priv->buffer));
    priv->buffer = GTK_TEXT_BUFFER(g_value_dup_object(value));
    c4s_custom_function_list_renew(list);
    break;
  case PROP_HIGHLIGHTER:
    if(priv->highlighter != NULL)
    {
      g_signal_handlers_disconnect_by_func(
        G_OBJECT(priv->highlighter),
        G_CALLBACK(c4s_custom_function_list_highlighted_cb),
        list
      );

      g_object_unref(G_OBJECT(priv->highlighter));
    }

    priv->highlighter = C4S_HIGHLIGHTER(g_value_dup_object(value));

    g_signal_connect_after(
      G_OBJECT(priv->highlighter),
      "highlighted",
      G_CALLBACK(c4s_custom_function_list_highlighted_cb),
      list
    );

    c4s_custom_function_list_renew(list);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_custom_function_list_get_property(GObject* object,
                                      guint prop_id,
                                      GValue* value,
                                      GParamSpec* pspec)
{
  C4sCustomFunctionList* list;
  C4sCustomFunctionListPrivate* priv;

  g_return_if_fail(C4S_IS_CUSTOM_FUNCTION_LIST(object));

  list = C4S_CUSTOM_FUNCTION_LIST(object);
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  switch(prop_id)
  {
  case PROP_BUFFER:
    g_value_set_object(value, priv->buffer);
    break;
  case PROP_HIGHLIGHTER:
    g_value_set_object(value, priv->highlighter);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_custom_function_list_function_added(C4sCustomFunctionList* list,
                                        C4sCustomFunction* function)
{
  C4sCustomFunctionListPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  g_object_ref(G_OBJECT(function));
  priv->function_list = g_list_prepend(priv->function_list, function);

  g_hash_table_insert(
    priv->function_table,
    /* no need to g_strdup since the function is valid while being
     * into the hash table. */
    (gpointer)c4s_function_get_name(C4S_FUNCTION(function)),
    function
  );
}

static void
c4s_custom_function_list_function_removed(C4sCustomFunctionList* list,
                                          C4sCustomFunction* function)
{
  C4sCustomFunctionListPrivate* priv;
  GList* item;

  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);

  g_hash_table_remove(
    priv->function_table,
    c4s_function_get_name(C4S_FUNCTION(function))
  );

  item = g_list_find(priv->function_list, function);
  g_assert(item != NULL);

  g_object_unref(G_OBJECT(item->data));
  priv->function_list = g_list_delete_link(priv->function_list, item);
}

static void
c4s_custom_function_list_class_init(C4sCustomFunctionListClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sCustomFunctionListPrivate));

  object_class->finalize = c4s_custom_function_list_finalize;
  object_class->set_property = c4s_custom_function_list_set_property;
  object_class->get_property = c4s_custom_function_list_get_property;
  
  klass->function_added = c4s_custom_function_list_function_added;
  klass->function_removed = c4s_custom_function_list_function_removed;

  g_object_class_install_property(
    object_class,
    PROP_BUFFER,
    g_param_spec_object(
      "buffer",
      "Text buffer",
      "GtkTextBuffer to scan for custom functions",
      GTK_TYPE_TEXT_BUFFER,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_HIGHLIGHTER,
    g_param_spec_object(
      "highlighter",
      "Syntax highlighter",
      "C4sHighlighter used to get information about the buffer's syntax",
      C4S_TYPE_HIGHLIGHTER,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  custom_function_list_signals[FUNCTION_ADDED] = g_signal_new(
    "function-added",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sCustomFunctionListClass, function_added),
    NULL, NULL,
    c4s_marshal_VOID__OBJECT,
    G_TYPE_NONE,
    1,
    C4S_TYPE_CUSTOM_FUNCTION
  );

  custom_function_list_signals[FUNCTION_REMOVED] = g_signal_new(
    "function-removed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sCustomFunctionListClass, function_removed),
    NULL, NULL,
    c4s_marshal_VOID__OBJECT,
    G_TYPE_NONE,
    1,
    C4S_TYPE_CUSTOM_FUNCTION
  );
}

GType
c4s_custom_function_list_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sCustomFunctionListClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_custom_function_list_class_init,
      NULL,
      NULL,
      sizeof(C4sCustomFunctionList),
      0,
      (GInstanceInitFunc)c4s_custom_function_list_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sCustomFunctionList",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sCustomFunctionList*
c4s_custom_function_list_new(GtkTextBuffer* buffer,
                             C4sHighlighter* highlighter)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_CUSTOM_FUNCTION_LIST,
    "buffer", buffer,
    "highlighter", highlighter,
    NULL
  );

  return C4S_CUSTOM_FUNCTION_LIST(object);
}

C4sCustomFunction*
c4s_custom_function_list_find_function(const C4sCustomFunctionList* list,
                                       const gchar* name)
{
  return C4S_CUSTOM_FUNCTION(
    g_hash_table_lookup(
      C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list)->function_table,
      name
    )
  );
}

void
c4s_custom_function_list_foreach(C4sCustomFunctionList* list,
                                 C4sCustomFunctionListForeachFunc func,
                                 gpointer user_data)
{
  C4sCustomFunctionListPrivate* priv;
  GList* item;

  priv = C4S_CUSTOM_FUNCTION_LIST_PRIVATE(list);
  for(item = priv->function_list; item != NULL; item = item->next)
    func(C4S_CUSTOM_FUNCTION(item->data), user_data);
}
