/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-enginefunction.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-fileutil.h>

#include <ctype.h>
#include <string.h>

static C4sFunctionClass* parent_class = NULL;

static const gchar SYNTAX_BEGIN[] = "<div class=\"fnsyntax\">";
static const gchar SYNTAX_END[] = "</div>";

static const gsize SYNTAX_BEGIN_LEN = sizeof(SYNTAX_BEGIN) - 1;
static const gsize SYNTAX_END_LEN = sizeof(SYNTAX_END) - 1;

static void
c4s_engine_function_advance_nonsense(const gchar** pos,
                                     const gchar* to)
{
  const gchar* i;
  const gchar* end;

  for(i = *pos; i < to; ++ i)
  {
    if(isspace(*i)) continue;

    if(*i == '<')
    {
      end = memchr(i + 1, '>', to - i);
      if(end == NULL)
      {
        *pos = to;
        return;
      }
      else
      {
        i = end;
      }
    }
    else
    {
      break;
    }
  }

  *pos = i;
}

static C4sValueType
c4s_engine_function_type_from_string(const gchar* from,
                                     gsize len)
{
  /* compatibility with unfamiliar / older types */
  if(c4s_substr_cmp(from, len, "BOOL", 4))
    return C4S_VALUE_TYPE_BOOL;
  else if(c4s_substr_cmp(from, len, "DWORD", 5))
    return C4S_VALUE_TYPE_INT;
  else if(c4s_substr_cmp(from, len, "C4ID", 4))
    return C4S_VALUE_TYPE_ID;
  else if(c4s_substr_cmp(from, len, "Any", 3))
    return C4S_VALUE_TYPE_ANY;
  else if(c4s_substr_cmp(from, len, "C4Value", 7))
    return C4S_VALUE_TYPE_ANY | C4S_VALUE_TYPE_REF;
  else
    return c4s_value_type_from_string(from, len);
}

static C4sValueType
c4s_engine_function_read_single_type(const gchar** pos,
                                     const gchar* to)
{
  C4sValueType type;
  const gchar* begin;
  const gchar* i;

  type = C4S_VALUE_TYPE_UNKNOWN;
  i = begin = *pos;

  while(isalnum(*i))
    ++ i;

  if(i != begin)
  {
    type = c4s_engine_function_type_from_string(begin, i - begin);
    if(type == C4S_VALUE_TYPE_UNKNOWN)
    {
      *pos = i;
      return type;
    }
  }

  c4s_engine_function_advance_nonsense(&i, to);

  /* * is for C4Value* */
  if(*i == '&' || *i == '*')
  {
    /* Only '&' is also a type and the same as 'any&' */
    if(type == C4S_VALUE_TYPE_UNKNOWN) type = C4S_VALUE_TYPE_ANY;
    type |= C4S_VALUE_TYPE_REF;

    if(*i == '&' && isalpha(*(i + 1)) )
      if(c4s_substr_cmp(i + 1, MIN(to - i - 1, 4), "amp;", 4))
        i += 4;

    ++ i;
    c4s_engine_function_advance_nonsense(&i, to);
  }

  *pos = i;
  return type;
}

static C4sValueType
c4s_engine_function_read_type(const gchar** pos,
                              const gchar* to)
{
  C4sValueType type;
  type = c4s_engine_function_read_single_type(pos, to);

  while(**pos == '/')
  {
    ++ *pos;
    c4s_engine_function_advance_nonsense(pos, to);

    type = c4s_engine_function_read_single_type(pos, to);
    if(type == C4S_VALUE_TYPE_UNKNOWN) return type;

    /* We do not handle multiple return types yet */
    type = C4S_VALUE_TYPE_ANY;
  }

  return type;
}

static gchar*
c4s_engine_function_read_name(const gchar** pos,
                              const gchar* to)
{
  const gchar* begin;
  gchar* name;

  begin = *pos;

  while(isalnum(**pos) || **pos == '_') ++ *pos;
  if(begin == *pos) return NULL;

  name = g_strndup(begin, *pos - begin);
  c4s_engine_function_advance_nonsense(pos, to);
  return name;
}

static void
c4s_engine_function_load_varargs(C4sEngineFunction* function,
                                 C4sValueType type)
{
  gsize params;
  gchar parbuf[5];

  c4s_function_get_parameters(C4S_FUNCTION(function), &params);

  for(; params < 10; ++ params)
  {
    sprintf(parbuf, "par%d", (int)params);
    c4s_function_add_parameter(
      C4S_FUNCTION(function),
      type,
      parbuf
    );
  }
}

static void
c4s_engine_function_class_init(C4sEngineFunctionClass* klass)
{
  parent_class = C4S_FUNCTION_CLASS(g_type_class_peek_parent(klass));
}

GType
c4s_engine_function_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sEngineFunctionClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_engine_function_class_init,
      NULL,
      NULL,
      sizeof(C4sEngineFunction),
      0,
      NULL,
      NULL
    };

    our_type = g_type_register_static(
      C4S_TYPE_FUNCTION,
      "C4sEngineFunction",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sEngineFunction*
c4s_engine_function_new(const gchar* doc_html,
                        gsize len)
{
  C4sEngineFunction* function;
  const gchar* syntax_begin;
  const gchar* syntax_end;
  const gchar* pos;

  C4sValueType type;
  gchar* name;

  syntax_begin = g_strstr_len(doc_html, len, SYNTAX_BEGIN);
  if(syntax_begin == NULL) return NULL;
  
  syntax_begin += SYNTAX_BEGIN_LEN;
  
  syntax_end = g_strstr_len(
    syntax_begin,
    len - (syntax_begin - doc_html),
    SYNTAX_END
  );

  if(syntax_end == NULL) return NULL;

  pos = syntax_begin;
  c4s_engine_function_advance_nonsense(&pos, syntax_end);

  type = c4s_engine_function_read_type(&pos, syntax_end);
  if(type == C4S_VALUE_TYPE_UNKNOWN) return NULL;

  name = c4s_engine_function_read_name(&pos, syntax_end);
  if(name == NULL) return NULL;

  if(*pos != '(')
  {
    g_free(name);
    return NULL;
  }

  ++ pos;
  c4s_engine_function_advance_nonsense(&pos, syntax_end);

  function = g_object_new(
    C4S_TYPE_ENGINE_FUNCTION,
    "return-type", type,
    "name", name,
    NULL
  );

  g_free(name);

  /* no parameters */
  if(*pos == ')') return function;

  for(;;)
  {
    if(strncmp(pos, "...", 3) == 0)
    {
      c4s_engine_function_load_varargs(function, C4S_VALUE_TYPE_ANY);
      break;
    }

    type = c4s_engine_function_read_type(&pos, syntax_end);
    if(type == C4S_VALUE_TYPE_UNKNOWN)
    {
      g_object_unref(function);
      return NULL;
    }

    if(strncmp(pos, "...", 3) == 0)
    {
      c4s_engine_function_load_varargs(function, type);
      break;
    }

    name = c4s_engine_function_read_name(&pos, syntax_end);
    if(name == NULL)
    {
      g_object_unref(function);
      return NULL;
    }

    c4s_function_add_parameter(
      C4S_FUNCTION(function),
      type,
      name
    );

    g_free(name);

    if(*pos == ')') break;
    if(*pos != ',')
    {
      g_object_unref(function);
      return NULL;
    }

    ++ pos;
    c4s_engine_function_advance_nonsense(&pos, syntax_end);
  }

  return function;
}

C4sEngineFunction*
c4s_engine_function_new_from_stream(FILE* stream,
                                    GError** error)
{
  C4sEngineFunction* function;
  C4sValueType type;
  guint32 param_count;
  guint32 i;
  guint8 byte;
  gchar* name;

  if(!c4s_file_read8(stream, &byte, error)) return NULL;
  type = byte;

  if(!c4s_file_read_string(stream, &name, error)) return NULL;

  function = g_object_new(
    C4S_TYPE_ENGINE_FUNCTION,
    "return-type", type,
    "name", name,
    NULL
  );

  g_free(name);

  if(!c4s_file_read8(stream, &byte, error))
  {
    g_object_unref(function);
    return NULL;
  }

  param_count = byte;

  for(i = 0; i < param_count; ++ i)
  {
    if(!c4s_file_read8(stream, &byte, error))
    {
      g_object_unref(function);
      return NULL;
    }

    type = byte;

    if(!c4s_file_read_string(stream, &name, error))
    {
      g_object_unref(function);
      return NULL;
    }

    c4s_function_add_parameter(
      C4S_FUNCTION(function),
      type,
      name
    );

    g_free(name);
  }

  return function;
}

gboolean
c4s_engine_function_save(const C4sEngineFunction* function,
                         FILE* stream,
                         GError** error)
{
  guint8 type;
  const gchar* name;
  const C4sFunctionParameter* params;
  gsize n_params;
  gsize n;

  type = c4s_function_get_return_type(C4S_FUNCTION(function));
  name = c4s_function_get_name(C4S_FUNCTION(function));
  params = c4s_function_get_parameters(C4S_FUNCTION(function), &n_params);

  if(!c4s_file_write8(stream, type, error)) return FALSE;
  if(!c4s_file_write_string(stream, name, error)) return FALSE;
  if(!c4s_file_write8(stream, n_params, error)) return FALSE;

  for(n = 0; n < n_params; ++ n)
  {
    type = params[n].type;
    if(!c4s_file_write8(stream, type, error))
      return FALSE;

    if(!c4s_file_write_string(stream, params[n].name, error))
      return FALSE;
  }

  return TRUE;
}
