/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-enginefunctionlist.h>
#include <c4scripter/c4s-functionquerydialog.h>
#include <c4scripter/c4s-functionprogressdialog.h>
#include <c4scripter/c4s-fileutil.h>
#include <c4scripter/c4s-marshal.h>

#include <c4scripter/loader/c4s-loader.h>

#include <gtk/gtkmessagedialog.h>
#include <gtk/gtktreemodel.h>

#include <errno.h>

enum {
  PROP_0,

  /* construct only */
  PROP_PARENT,
  PROP_PREFERENCES
};

enum {
  RELOADED,
  LAST_SIGNAL
};

typedef struct _C4sEngineFunctionListPrivate C4sEngineFunctionListPrivate;
struct _C4sEngineFunctionListPrivate {
  /* Loading stuff */
  const C4sLoaderBackendFuncs* backend_funcs;
  C4sLoaderBackend* backend;
  unsigned int current_function;
  unsigned int function_count;
  GSList* new_functions; /* Functions currently being loaded */

  /* Properties */
  C4sPreferences* preferences;
  GtkWindow* parent;

  /* Function table */
  GHashTable* hash_table; /* fast lookup by function name */
  GtkListStore* list_store;

  /* UI */
  C4sFunctionQueryDialog* query_dialog;
  C4sFunctionProgressDialog* progress_dialog;
};

#define C4S_ENGINE_FUNCTION_LIST_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_ENGINE_FUNCTION_LIST, \
    C4sEngineFunctionListPrivate \
  ))

static GObjectClass* parent_class = NULL;
static guint enginefunctionlist_signals[LAST_SIGNAL];

static const guint8 CURRENT_FUNC_FILE_VERSION = 1;

static void
c4s_engine_function_list_destroy_func(gpointer data,
                                      G_GNUC_UNUSED gpointer user_data)
{
  /* data is C4sEngineFunction */
  g_object_unref(data);
}

static void
c4s_engine_function_list_clear_new_functions(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

	/* Remove already read functions */
  g_slist_foreach(
    priv->new_functions,
    c4s_engine_function_list_destroy_func,
    NULL
  );

  g_slist_free(priv->new_functions);
  priv->new_functions = NULL;
}

static void
c4s_engine_function_list_error_response_cb(GtkMessageDialog* dialog,
                                           G_GNUC_UNUSED int response_id,
                                           G_GNUC_UNUSED gpointer user_data)
{
  gtk_widget_destroy(GTK_WIDGET(dialog));
}

static void
c4s_engine_function_list_loader_close(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  priv->backend_funcs->deinit(priv->backend);
  priv->backend_funcs = NULL;
  priv->backend = NULL;

  priv->function_count = 0;
  priv->current_function = 0;

  gtk_widget_destroy(GTK_WIDGET(priv->progress_dialog));

  c4s_engine_function_list_clear_new_functions(list);
}

static void
c4s_engine_function_list_loader_show_error(C4sEngineFunctionList* list,
                                           const GError* error)
{
  C4sEngineFunctionListPrivate* priv;
  GtkWidget* dialog;

  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  dialog = gtk_message_dialog_new(
    priv->parent,
    GTK_DIALOG_DESTROY_WITH_PARENT,
    GTK_MESSAGE_ERROR,
    GTK_BUTTONS_OK,
    "An error has occured while loading the function list"
  );

  gtk_message_dialog_format_secondary_text(
    GTK_MESSAGE_DIALOG(dialog),
    "%s",
    error->message
  );
  
  gtk_window_set_title(
    GTK_WINDOW(dialog),
    "Error loading function list"
  );

  g_signal_connect(
    G_OBJECT(dialog),
    "response",
    G_CALLBACK(c4s_engine_function_list_error_response_cb),
    list
  );

  gtk_widget_show(dialog);
}

static void
c4s_engine_function_list_loader_init_done(unsigned int function_count,
                                          gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);
  
  priv->function_count = function_count;
  priv->current_function = 0;

  c4s_function_progress_dialog_set_progress(priv->progress_dialog, 0.0);

  if(priv->function_count == 0)
  {
    /* Nothing to do, just clear current functions and close loader */
    g_signal_emit(list, enginefunctionlist_signals[RELOADED], 0);
    c4s_engine_function_list_loader_close(list);
  }
  else
  {
    priv->backend_funcs->load_function(priv->backend, priv->current_function);
  }
}

static void
c4s_engine_function_list_loader_init_failed(const GError* error,
                                            gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  c4s_engine_function_list_loader_show_error(list, error);
  c4s_engine_function_list_loader_close(list);
}

static void
c4s_engine_function_list_loader_load_next(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  ++ priv->current_function;

  c4s_function_progress_dialog_set_progress(
    priv->progress_dialog,
    (double)priv->current_function / (double)priv->function_count
  );
  
  if(priv->current_function == priv->function_count)
  {
    /* Loaded all functions */
    g_signal_emit(list, enginefunctionlist_signals[RELOADED], 0);
    c4s_engine_function_list_loader_close(list);
  }
  else
  {
    priv->backend_funcs->load_function(priv->backend, priv->current_function);
  }
}

static void
c4s_engine_function_list_loader_function_done(C4sEngineFunction* function,
                                              gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  priv->new_functions = g_slist_prepend(priv->new_functions, function);
  c4s_engine_function_list_loader_load_next(list);
}

static void
c4s_engine_function_list_loader_function_failed(const gchar* name,
                                                const GError* error,
                                                gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  switch(error->code)
  {
  case C4S_LOADER_TRANSPORT_ERROR:
    /* Transport broken */
    c4s_engine_function_list_loader_show_error(list, error);
    c4s_engine_function_list_loader_close(list);
    break;
  case C4S_LOADER_PARSE_ERROR:
    /* Error is only local to function, try next one */
    printf("Omitted %s: %s\n", name, error->message);
    c4s_engine_function_list_loader_load_next(list);
    break;
  }
}

static void
c4s_engine_function_list_loader_set_status(const gchar* status,
                                           gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  c4s_function_progress_dialog_set_status(priv->progress_dialog, status);
}

static gboolean
c4s_engine_function_list_write_to_file(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  C4sEngineFunction* function;
  GtkTreeModel* model;
  GtkTreeIter iter;
  gboolean result;

  FILE* stream;
  GError* error;
  gchar* dirname;
  gchar* filename;
  guint32 function_count;
  
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);
  
  error = NULL;
  dirname = g_build_filename(g_get_home_dir(), ".c4scripter", NULL);
  filename = g_build_filename(dirname, "functions", NULL);

  g_mkdir_with_parents(dirname, 0755);
  stream = fopen(filename, "wb");
  
  g_free(dirname);
  g_free(filename);

  if(stream == NULL)
  {
    g_set_error(
      &error,
      g_quark_from_static_string("G_FILE_ERROR"),
      g_file_error_from_errno(errno),
      "%s",
      g_strerror(errno)
    );
    
    goto write_error;
  }

  if(c4s_file_write8(stream, CURRENT_FUNC_FILE_VERSION, &error) == FALSE)
    goto write_error;

  model = GTK_TREE_MODEL(priv->list_store);
  function_count = gtk_tree_model_iter_n_children(model, NULL);

  if(c4s_file_write32(stream, function_count, &error) == FALSE)
    goto write_error;

  if(gtk_tree_model_get_iter_first(model, &iter) == TRUE)
  {
    do
    {
      gtk_tree_model_get(model, &iter, 1, &function, -1);
      result = c4s_engine_function_save(function, stream, &error);
      g_object_unref(G_OBJECT(function));
      if(result == FALSE) goto write_error;
    } while(gtk_tree_model_iter_next(model, &iter) == TRUE);
  }
  
  fclose(stream);
  return TRUE;

write_error:
  if(stream != NULL)
    fclose(stream);

  c4s_engine_function_list_loader_show_error(list, error);
  g_error_free(error);
  return FALSE;
}

static gboolean
c4s_engine_function_list_read_from_file(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  gchar* func_file;
  FILE* stream;

  C4sEngineFunction* function;
  guint8 version_flag;
  guint32 func_count;
  guint32 i;

  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  func_file = g_build_filename(g_get_home_dir(), ".c4scripter", "functions", NULL);

  stream = fopen(func_file, "rb");
  g_free(func_file);
  
  if(stream == NULL) return FALSE;

  /* One of the rare cases where goto is actually useful... */
  if(c4s_file_read8(stream, &version_flag, NULL) == FALSE) goto read_error;
  if(version_flag != CURRENT_FUNC_FILE_VERSION) goto read_error;
  
  if(c4s_file_read32(stream, &func_count, NULL) == FALSE) goto read_error;

  for(i = 0; i < func_count; ++ i)
  {
    function = c4s_engine_function_new_from_stream(stream, NULL);
    if(function == NULL) goto read_error;

    priv->new_functions = g_slist_prepend(priv->new_functions, function);
  }
  
  fclose(stream);
  g_signal_emit(list, enginefunctionlist_signals[RELOADED], 0);
  return TRUE;

read_error:
  fclose(stream);
  c4s_engine_function_list_clear_new_functions(list);
  return FALSE;
}

static void
c4s_engine_function_list_reloaded(C4sEngineFunctionList* list)
{
  C4sEngineFunctionListPrivate* priv;
  GtkTreeIter iter;
  GSList* item;

  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  /* This should just clear the hash table, but there is
   * no g_hash_table_clear */
  g_hash_table_destroy(priv->hash_table);
  priv->hash_table = g_hash_table_new_full(
    g_str_hash,
    g_str_equal,
    NULL, /* owned by function */
    NULL /* function is not ref'ed */
  );

  gtk_list_store_clear(priv->list_store);

  for(item = priv->new_functions; item != NULL; item = item->next)
  {
    gtk_list_store_prepend(priv->list_store, &iter);
    gtk_list_store_set(
      priv->list_store,
      &iter,
      0, c4s_function_get_name(C4S_FUNCTION(item->data)),
      1, C4S_ENGINE_FUNCTION(item->data),
      -1
    );
    
    g_hash_table_insert(
      priv->hash_table,
      /* no need to g_strdup since the function is valid while being
       * into the hash table. */
      (gpointer)c4s_function_get_name(C4S_FUNCTION(item->data)),
      item->data
    );
  }

  c4s_engine_function_list_clear_new_functions(list);
  c4s_engine_function_list_write_to_file(list);
}

static void
c4s_engine_function_list_progress_response_cb(G_GNUC_UNUSED GtkDialog* dialog,
                                              int response_id,
                                              gpointer user_data)
{
  C4sEngineFunctionList* list;
  list = C4S_ENGINE_FUNCTION_LIST(user_data);

  if(response_id == GTK_RESPONSE_CANCEL ||
     response_id == GTK_RESPONSE_DELETE_EVENT)
  {
    /* Cancel loading when progress dialog is closed */
    c4s_engine_function_list_loader_close(list);
  }
}

static void
c4s_engine_function_list_progress_destroy_cb(G_GNUC_UNUSED GtkDialog* dialog,
                                             gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);
  
  priv->progress_dialog = NULL;
}

static void
c4s_engine_function_list_load_functions(C4sEngineFunctionList* list)
{
  static const C4sLoaderBackendCallbacks BACKEND_CALLBACKS = {
    c4s_engine_function_list_loader_init_done,
    c4s_engine_function_list_loader_init_failed,
    c4s_engine_function_list_loader_function_done,
    c4s_engine_function_list_loader_function_failed,
    c4s_engine_function_list_loader_set_status,
  };

  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  if(priv->query_dialog != NULL) return;
  if(priv->backend != NULL) return;

  priv->progress_dialog = c4s_function_progress_dialog_new(
    priv->parent,
    GTK_DIALOG_NO_SEPARATOR
  );

  g_signal_connect(
    G_OBJECT(priv->progress_dialog),
    "response",
    G_CALLBACK(c4s_engine_function_list_progress_response_cb),
    list
  );
    
  g_signal_connect(
    G_OBJECT(priv->progress_dialog),
    "destroy",
    G_CALLBACK(c4s_engine_function_list_progress_destroy_cb),
    list
  );

  switch(c4s_preferences_get_function_query_action(priv->preferences))
  {
  case C4S_FUNCTION_QUERY_ACTION_CHM:
    priv->backend_funcs = &C4S_LOADER_BACKEND_CHM;
    priv->backend = priv->backend_funcs->init(
      priv->preferences,
      &BACKEND_CALLBACKS,
      list
    );

    break;
  case C4S_FUNCTION_QUERY_ACTION_ONLINE:
    priv->backend_funcs = &C4S_LOADER_BACKEND_ONLINE;
    priv->backend = priv->backend_funcs->init(
      priv->preferences,
      &BACKEND_CALLBACKS,
      list
    );

    break;
  default:
    g_assert_not_reached();
    break;
  }

  gtk_widget_show(GTK_WIDGET(priv->progress_dialog));
}

static void
c4s_engine_function_list_query_response_cb(G_GNUC_UNUSED
                                           C4sFunctionQueryDialog* dialog,
                                           int response_id,
                                           gpointer user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;
  
  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);
  
  if(response_id == GTK_RESPONSE_ACCEPT)
    c4s_engine_function_list_load_functions(list);
  
  if(response_id != GTK_RESPONSE_DELETE_EVENT)
    gtk_widget_destroy(GTK_WIDGET(priv->query_dialog));
}

static void
c4s_engine_function_list_query_destroy_cb(G_GNUC_UNUSED
                                          C4sFunctionQueryDialog* dialog,
                                          void* user_data)
{
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;

  list = C4S_ENGINE_FUNCTION_LIST(user_data);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);

  priv->query_dialog = NULL;
}

static void
c4s_engine_function_list_init(C4sEngineFunctionList* engine_function_list)
{
  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(engine_function_list);

  priv->backend_funcs = NULL;
  priv->backend = NULL;
  priv->current_function = 0;
  priv->function_count = 0;
  priv->new_functions = NULL;
  
  priv->preferences = NULL;
  priv->parent = NULL;
  
  priv->list_store = gtk_list_store_new(
    2,
    G_TYPE_STRING,
    C4S_TYPE_ENGINE_FUNCTION
  );
  
  priv->hash_table = g_hash_table_new_full(
    g_str_hash,
    g_str_equal,
    NULL, /* owned by function */
    NULL /* function is not ref'ed */
  );
  
  priv->query_dialog = NULL;
  priv->progress_dialog = NULL;
}

static void 
c4s_engine_function_list_finalize(GObject* object)
{
  C4sEngineFunctionList* engine_function_list;
  C4sEngineFunctionListPrivate* priv;

  engine_function_list = C4S_ENGINE_FUNCTION_LIST(object);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(engine_function_list);

  if(priv->backend != NULL)
    priv->backend_funcs->deinit(priv->backend);

  c4s_engine_function_list_clear_new_functions(engine_function_list);

  g_object_unref(G_OBJECT(priv->preferences));

  g_hash_table_destroy(priv->hash_table);
  g_object_unref(G_OBJECT(priv->list_store));
  
  if(priv->query_dialog != NULL)
    gtk_widget_destroy(GTK_WIDGET(priv->query_dialog));
  if(priv->progress_dialog != NULL)
    gtk_widget_destroy(GTK_WIDGET(priv->progress_dialog));

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_engine_function_list_set_property(GObject* object,
                                      guint prop_id,
                                      const GValue* value,
                                      GParamSpec* pspec)
{
  C4sEngineFunctionList* engine_function_list;
  C4sEngineFunctionListPrivate* priv;

  g_return_if_fail(C4S_IS_ENGINE_FUNCTION_LIST(object));

  engine_function_list = C4S_ENGINE_FUNCTION_LIST(object);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(engine_function_list);

  switch(prop_id)
  {
  case PROP_PARENT:
    priv->parent = g_value_get_object(value);
    break;
  case PROP_PREFERENCES:
    if(priv->preferences != NULL) g_object_unref(priv->preferences);
    priv->preferences = C4S_PREFERENCES(g_value_dup_object(value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_engine_function_list_get_property(GObject* object,
                                      guint prop_id,
                                      G_GNUC_UNUSED GValue* value,
                                      GParamSpec* pspec)
{
  C4sEngineFunctionList* engine_function_list;
  C4sEngineFunctionListPrivate* priv;

  g_return_if_fail(C4S_IS_ENGINE_FUNCTION_LIST(object));

  engine_function_list = C4S_ENGINE_FUNCTION_LIST(object);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(engine_function_list);

  switch(prop_id)
  {
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_engine_function_list_class_init(C4sEngineFunctionListClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sEngineFunctionListPrivate));

  object_class->finalize = c4s_engine_function_list_finalize;
  object_class->set_property = c4s_engine_function_list_set_property;
  object_class->get_property = c4s_engine_function_list_get_property;
  klass->reloaded = c4s_engine_function_list_reloaded;

  g_object_class_install_property(
    object_class,
    PROP_PARENT,
    g_param_spec_object(
      "parent",
      "Parent",
      "Window used as parent for several dialogs",
      GTK_TYPE_WINDOW,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "Preferences to read additional information from",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  enginefunctionlist_signals[RELOADED] = g_signal_new(
    "reloaded",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sEngineFunctionListClass, reloaded),
    NULL, NULL,
    c4s_marshal_VOID__VOID,
    G_TYPE_NONE,
    0
  );
}

GType
c4s_engine_function_list_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sEngineFunctionListClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_engine_function_list_class_init,
      NULL,
      NULL,
      sizeof(C4sEngineFunctionList),
      0,
      (GInstanceInitFunc)c4s_engine_function_list_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sEngineFunctionList",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sEngineFunctionList*
c4s_engine_function_list_new(GtkWindow* parent,
                             C4sPreferences* preferences)
{
  GObject* object;
  C4sEngineFunctionList* list;
  C4sEngineFunctionListPrivate* priv;

  object = g_object_new(
    C4S_TYPE_ENGINE_FUNCTION_LIST,
    "parent", parent,
    "preferences", preferences,
    NULL
  );
  
  list = C4S_ENGINE_FUNCTION_LIST(object);
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);
  
  if(c4s_engine_function_list_read_from_file(list) == FALSE)
  {
    priv->query_dialog = c4s_function_query_dialog_new(
      priv->parent,
      GTK_DIALOG_NO_SEPARATOR,
      priv->preferences
    );
    
    g_signal_connect(
      G_OBJECT(priv->query_dialog),
      "response",
      G_CALLBACK(c4s_engine_function_list_query_response_cb),
      list
    );
    
    g_signal_connect(
      G_OBJECT(priv->query_dialog),
      "destroy",
      G_CALLBACK(c4s_engine_function_list_query_destroy_cb),
      list
    );
    
    gtk_widget_show(GTK_WIDGET(priv->query_dialog));
  }

  return list;
}

const C4sEngineFunction*
c4s_engine_function_list_find_function(const C4sEngineFunctionList* list,
                                       const gchar* name)
{
  C4sEngineFunctionListPrivate* priv;
  priv = C4S_ENGINE_FUNCTION_LIST_PRIVATE(list);  

  return C4S_ENGINE_FUNCTION(g_hash_table_lookup(priv->hash_table, name));
}

GtkListStore*
c4s_engine_function_list_get_list_store(const C4sEngineFunctionList* list)
{
  return C4S_ENGINE_FUNCTION_LIST_PRIVATE(list)->list_store;
}

void
c4s_engine_function_list_reload(C4sEngineFunctionList* list)
{
  c4s_engine_function_list_load_functions(list);
}
