/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-highlighter.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-marshal.h>

#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextmark.h>

#include <glib/gunicode.h>

#include <string.h>

enum {
  PROP_0,

  /* TODO: Consider making these construct only */
  PROP_BUFFER,
  PROP_KEYWORDS,
  PROP_ENGINE_FUNCTION_LIST,
  PROP_PREFERENCES
};

enum {
  HIGHLIGHTED = 0,
  LAST_SIGNAL
};

typedef struct _C4sHighlighterPrivate C4sHighlighterPrivate;
struct _C4sHighlighterPrivate {
        GtkTextBuffer* buffer;
        const gchar** keywords;
        C4sEngineFunctionList* engine_function_list;
        C4sPreferences* preferences;

        /* primary tags: Tags covering ranges of text delimited by
         * character sequences. Primary tags have priority over secondary
         * tags */
        GtkTextTag* tag_string;
        GtkTextTag* tag_single_comment;
        GtkTextTag* tag_multi_comment;

        /* secondary tags: Tags applied to single identifiers. */
        GtkTextTag* tag_keyword;
        GtkTextTag* tag_function;
        GtkTextTag* tag_number;

        /* helper tag, is applied to all ranges where either
         * tag_single_comment or tag_multi_comment is applied to. */
        GtkTextTag* tag_comment;

        gboolean is_user_action;
        GtkTextMark* mark_user_action;
};

#define C4S_HIGHLIGHTER_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_HIGHLIGHTER, \
    C4sHighlighterPrivate \
  ))

static GObjectClass* parent_class;
static guint highlighter_signals[LAST_SIGNAL];

static gboolean
c4s_highlighter_predicate_number_cb(gunichar c,
                                    G_GNUC_UNUSED gpointer user_data)
{
  return g_unichar_isdigit(c);
}

static gboolean
c4s_highlighter_predicate_nonnumber_cb(gunichar c,
                                       G_GNUC_UNUSED gpointer user_data)
{
  return !g_unichar_isdigit(c);
}

static gboolean
c4s_highlighter_is_escaped(const GtkTextIter* iter)
{
  gboolean escaped;
  GtkTextIter copy_iter;

  escaped = FALSE;
  copy_iter = *iter;

  while(gtk_text_iter_backward_char(&copy_iter) == TRUE)
  {
    if(gtk_text_iter_get_char(&copy_iter) == '\\')
      escaped = !escaped;
    else
      break;
  }

  return escaped;
}

/* Searches tags for a primary highlight tag */
static GtkTextTag*
c4s_highlighter_get_highlight_tag(const C4sHighlighter* highlighter,
                                  const GSList* tags)
{
  C4sHighlighterPrivate* priv;
  const GSList* item;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  for(item = tags; item != NULL; item = item->next)
  {
    if(item->data == priv->tag_string)
      return priv->tag_string;
    else if(item->data == priv->tag_single_comment)
      return priv->tag_single_comment;
    else if(item->data == priv->tag_multi_comment)
      return priv->tag_multi_comment;
  }

  return NULL;
}

/* Returns the primary highlight tag with which the position at iter
 * is highlighted */
static GtkTextTag*
c4s_highlighter_get_highlight_at(const C4sHighlighter* highlighter,
                                 const GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextIter local_iter;
  GSList* tags;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  local_iter = *iter;

  /* Default highlight mode at start */
  if(gtk_text_iter_backward_char(&local_iter) == FALSE)
    return NULL;

  tags = gtk_text_iter_get_tags(&local_iter);
  tag = c4s_highlighter_get_highlight_tag(highlighter, tags);
  g_slist_free(tags);

  if(tag == priv->tag_string)
  {
    /* local_iter may be terminating '"', so highlight mode falls back to
     * default (since iter is one character past local_iter), but not on
     * starting '"'. */
    if((gtk_text_iter_get_char(&local_iter) == '\"') &&
       (!c4s_highlighter_is_escaped(&local_iter)) &&
       !gtk_text_iter_begins_tag(&local_iter, tag))
    {
      return NULL;
    }
    else
    {
      return tag;
    }
  }
  else if(tag == priv->tag_multi_comment)
  {
    /* iter may be past comment end, so highlight mode falls back to
     * default */
    if(gtk_text_iter_get_char(&local_iter) != '/') return tag;

    gtk_text_iter_backward_char(&local_iter);
    if(gtk_text_iter_get_char(&local_iter) != '*') return tag;

    return NULL;
  }
  else
  {
    return tag;
  }
}

static void
c4s_highlighter_apply_primary_tag(C4sHighlighter* highlighter,
                                  GtkTextTag* tag,
                                  const GtkTextIter* begin,
                                  const GtkTextIter* end)
{
  C4sHighlighterPrivate* priv;
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  gtk_text_buffer_remove_all_tags(priv->buffer, begin, end);
  gtk_text_buffer_apply_tag(priv->buffer, tag, begin, end);

  if(tag == priv->tag_single_comment || tag == priv->tag_multi_comment)
    gtk_text_buffer_apply_tag(priv->buffer, priv->tag_comment, begin, end);
}

static void
c4s_highlighter_highlight_keywords(C4sHighlighter* highlighter,
                                   const GtkTextIter* begin,
                                   const GtkTextIter* end)
{
  C4sHighlighterPrivate* priv;
  const gchar** keyword;

  GtkTextIter match_begin;
  GtkTextIter match_end;
  GtkTextIter search_begin;
  gboolean result;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  if(priv->keywords == NULL) return;

  for(keyword = priv->keywords; *keyword != NULL; ++ keyword)
  {
    search_begin = *begin;

    for(;;)
    {
      result = c4s_forward_search(
        &search_begin,
        *keyword,
        0,
        &match_begin,
        &match_end,
        end
      );

      if(result == FALSE) break;

      search_begin = match_end;
      if(c4s_starts_identifier(&match_begin) &&
         c4s_ends_identifier(&match_end))
      {
        gtk_text_buffer_apply_tag(
          priv->buffer,
          priv->tag_keyword,
          &match_begin,
          &match_end
        );
      }
    }
  }
}

static void
c4s_highlighter_highlight_functions(C4sHighlighter* highlighter,
                                    const GtkTextIter* begin,
                                    const GtkTextIter* end)
{
  C4sHighlighterPrivate* priv;
  C4sEngineFunction* function;
  GtkListStore* store;
  GtkTreeIter iter;

  GtkTextIter match_begin;
  GtkTextIter match_end;
  GtkTextIter search_begin;
  gboolean result;
  
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  if(priv->engine_function_list == NULL) return;

  store = c4s_engine_function_list_get_list_store(priv->engine_function_list);
  if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter) == FALSE)
    return;

  do
  {
    gtk_tree_model_get(
      GTK_TREE_MODEL(store),
      &iter,
      1,
      &function,
      -1
    );

    search_begin = *begin;
    for(;;)
    {
      result = c4s_forward_search(
        &search_begin,
        c4s_function_get_name(C4S_FUNCTION(function)),
        0,
        &match_begin,
        &match_end,
        end
      );
      
      if(result == FALSE) break;
      
      search_begin = match_end;
      if(c4s_starts_identifier(&match_begin) &&
         c4s_ends_identifier(&match_end))
      {
        gtk_text_buffer_apply_tag(
          priv->buffer,
          priv->tag_function,
          &match_begin,
          &match_end
        );
      }
    }
  } while(gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter) == TRUE);
}

static void
c4s_highlighter_highlight_numbers(C4sHighlighter* highlighter,
                                  const GtkTextIter* begin,
                                  const GtkTextIter* end)
{
  C4sHighlighterPrivate* priv;
  GtkTextIter match_begin;
  GtkTextIter match_end;
  gboolean result;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  match_begin = *begin;

  for(;;)
  {
    if(!g_unichar_isdigit(gtk_text_iter_get_char(&match_begin)))
    {
      result = gtk_text_iter_forward_find_char(
        &match_begin,
        c4s_highlighter_predicate_number_cb,
        NULL,
        end
      );

      if(result == FALSE) break;
    }

    match_end = match_begin;

    gtk_text_iter_forward_find_char(
      &match_end,
      c4s_highlighter_predicate_nonnumber_cb,
      NULL,
      end
    );

    if(c4s_starts_identifier(&match_begin) && c4s_ends_identifier(&match_end))
    {
      gtk_text_buffer_apply_tag(
        priv->buffer,
        priv->tag_number,
        &match_begin,
        &match_end
      );
    }

    match_begin = match_end;
  }
}

static void
c4s_highlighter_highlighted_cb(C4sHighlighter* highlighter,
                               GtkTextIter* begin,
                               GtkTextIter* end)
{
  /* This function moves from and to to locations where actual rehighlighting
   * has been performed. */
  C4sHighlighterPrivate* priv;
  GtkTextTag* cur_tag;
  GtkTextIter prev;
  GtkTextIter iter;
  gunichar c;

  GSList* tags;
  GtkTextTag* tag;
  GtkTextIter range_begin;
  GtkTextIter range_end;
  GtkTextIter limit;
  gboolean exceed_end;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  if(priv->buffer == NULL) return;

  prev = *begin;
  c4s_backward_identifier(begin);

  /* When the iterator has not been moved, this means that begin is not
   * located at an identifier */
  if(gtk_text_iter_compare(&prev, begin) == 0)
  {
    /* Move out of comment delimiter or even some more complicated sequences
     * of '/' and '*' */
    while(gtk_text_iter_backward_char(begin) == TRUE)
    {
      c = gtk_text_iter_get_char(begin);
      if(c != '*' && c != '/')
      {
        gtk_text_iter_forward_char(begin);
        break;
      }
    }
  }

  c4s_forward_identifier(end);
  cur_tag = c4s_highlighter_get_highlight_at(highlighter, begin);

  /* whether the original end location has already been exceeded */
  exceed_end = FALSE;

  for(iter = *begin; gtk_text_iter_is_end(&iter) == FALSE;)
  {
    if(gtk_text_iter_compare(&iter, end) >= 0)
    {
      /* Reached end of range to re-highlight. It is possible that the
       * primary highlight tag at this position has changed, for example
       * if someone removed a '/' belonging to a '//' below the original
       * rehighlight range. We have to rehighlight stuff further until
       * we reach a position where the primary highlight tag does not
       * differ anymore. */
      tags = gtk_text_iter_get_tags(&iter);
      tag = c4s_highlighter_get_highlight_tag(highlighter, tags);
      g_slist_free(tags);

      if(cur_tag == tag) break;

      exceed_end = TRUE;
    }

    if(cur_tag == NULL)
    {
      range_begin = iter;

      while(gtk_text_iter_is_end(&iter) == FALSE)
      {
        /* exceeded end during rehighlight, maybe we can stop here */
        if(exceed_end == FALSE && gtk_text_iter_compare(&iter, end) >= 0)
        {
          range_end = iter;
          break;
        }

        c = gtk_text_iter_get_char(&iter);
        if(c == '\"')
        {
          /* String begins */
          range_end = iter;
          gtk_text_iter_forward_char(&iter);

          /* Highlighting leading " */
          cur_tag = priv->tag_string;

          c4s_highlighter_apply_primary_tag(
            highlighter,
            cur_tag,
            &range_end,
            &iter
          );

          break;
        }
        else if(c == '/')
        {
          /* comment begins */
          gtk_text_iter_forward_char(&iter);
          c = gtk_text_iter_get_char(&iter);

          if(c == '*' || c == '/')
          {
            gtk_text_iter_forward_char(&iter);
            range_end = iter;
            gtk_text_iter_backward_chars(&range_end, 2);

            cur_tag = priv->tag_multi_comment;
            if(c == '/') cur_tag = priv->tag_single_comment;

            c4s_highlighter_apply_primary_tag(
              highlighter,
              cur_tag,
              &range_end,
              &iter
            );

            break;
          }
        }
        else
        {
          gtk_text_iter_forward_char(&iter);
        }
      }

      /* Note that range_end is already set if another primary highlight
       * has been found */
      if(cur_tag == NULL) range_end = iter;

      /* No primary highlight between range_begin and range_end, so this is
       * neither comment nor string. Now check for secondary highlight in
       * this region */
      gtk_text_buffer_remove_all_tags(
        priv->buffer,
        &range_begin,
        &range_end
      );

      c4s_highlighter_highlight_keywords(
        highlighter,
        &range_begin,
        &range_end
      );

      c4s_highlighter_highlight_numbers(
        highlighter,
        &range_begin,
        &range_end
      );
      
      c4s_highlighter_highlight_functions(
        highlighter,
        &range_begin,
        &range_end
      );
    }
    else if(cur_tag == priv->tag_single_comment)
    {
      if(!gtk_text_iter_ends_line(&iter))
      {
        range_begin = iter;
        gtk_text_iter_forward_to_line_end(&iter);

        cur_tag = NULL;
        c4s_highlighter_apply_primary_tag(
          highlighter,
          priv->tag_single_comment,
          &range_begin,
          &iter
        );
      }

      /* continue at next line */
      gtk_text_iter_forward_line(&iter);
      cur_tag = NULL;
    }
    else if(cur_tag == priv->tag_multi_comment)
    {
      range_begin = iter;

      gtk_text_buffer_get_end_iter(priv->buffer, &limit);
      if(exceed_end == FALSE) limit = *end;

      if(!c4s_forward_search(&iter, "*/", 0, NULL, &range_end, &limit))
        range_end = limit;
      else
        cur_tag = NULL;

      iter = range_end;
      c4s_highlighter_apply_primary_tag(
        highlighter,
        priv->tag_multi_comment,
        &range_begin,
        &range_end
      );
    }
    else if(cur_tag == priv->tag_string)
    {
      range_begin = limit = iter;
      if(!gtk_text_iter_ends_line(&limit)) gtk_text_iter_forward_line(&limit);

      if(!c4s_forward_search(&iter, "\"", 0, NULL, &range_end, &limit))
        range_end = limit;

      iter = range_end;

      c4s_highlighter_apply_primary_tag(
        highlighter,
        priv->tag_string,
        &range_begin,
        &range_end
      );

      if(gtk_text_iter_ends_line(&iter))
      {
        cur_tag = NULL;
      }
      else
      {
        /* TODO: Find with predicate to omit extra escaped check */
        gtk_text_iter_backward_char(&range_end);
        if(!c4s_highlighter_is_escaped(&range_end))
          cur_tag = NULL;
      }
    }
  }

  /* Adjust end iter to really rehighlighted region */
  *end = iter;
}

static void
c4s_highlighter_rehighlight(C4sHighlighter* highlighter,
                            const GtkTextIter* from,
                            const GtkTextIter* to)
{
  C4sHighlighterPrivate* priv;
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  if(priv->buffer == NULL) return;

  g_signal_emit(
    G_OBJECT(highlighter),
    highlighter_signals[HIGHLIGHTED],
    0,
    from,
    to
  );
}

static void
c4s_highlighter_rehighlight_functions(C4sHighlighter* highlighter)
{
  /* TODO: Do only scan non-comment and non-string areas by calling
   * c4s_highlighter_highlight_functions as appropriate */
  C4sHighlighterPrivate* priv;
  GtkTextIter start;
  GtkTextIter end;
  
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  
  gtk_text_buffer_get_start_iter(priv->buffer, &start);
  gtk_text_buffer_get_end_iter(priv->buffer, &end);
  
  c4s_highlighter_rehighlight(highlighter, &start, &end);
}

static void
c4s_highlighter_recolor(C4sHighlighter* highlighter)
{
  C4sHighlighterPrivate* priv;
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  if(priv->buffer == NULL || priv->preferences == NULL) return;

  g_object_set(
    G_OBJECT(priv->tag_string),
    "foreground-gdk",
    c4s_preferences_get_color_string(priv->preferences),
    NULL
  );

  g_object_set(
    G_OBJECT(priv->tag_single_comment),
    "foreground-gdk",
    c4s_preferences_get_color_comment(priv->preferences),
    NULL
  );

  g_object_set(
    G_OBJECT(priv->tag_multi_comment),
    "foreground-gdk",
    c4s_preferences_get_color_comment(priv->preferences),
    NULL
  );

  g_object_set(
    G_OBJECT(priv->tag_keyword),
    "foreground-gdk",
    c4s_preferences_get_color_keyword(priv->preferences),
    NULL
  );
  
  g_object_set(
    G_OBJECT(priv->tag_function),
    "foreground-gdk",
    c4s_preferences_get_color_function(priv->preferences),
    NULL
  );

  g_object_set(
    G_OBJECT(priv->tag_number),
    "foreground-gdk",
    c4s_preferences_get_color_number(priv->preferences),
    NULL
  );
}

static void
c4s_highlighter_insert_cb(G_GNUC_UNUSED GtkTextBuffer* buffer,
                          GtkTextIter* pos,
                          gchar* text,
                          gint len,
                          gpointer user_data)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;
  GtkTextIter begin;

  highlighter = C4S_HIGHLIGHTER(user_data);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  begin = *pos;

  gtk_text_iter_backward_chars(&begin, g_utf8_strlen(text, len));

  /* If this is a user action, we assume that text is inserted at the same
   * position within one user action and multiple insert-text signal
   * emissions. This allows us to optimize this case by storing the position
   * where the text insertion began (priv->mark_user_action) and only
   * rehighlighting once in end_user_action */

  /* TODO: Note that this behaviour fails if text is inserted at different
   * locations within a single user action. To correct this, we should keep
   * a list of GtkTextMark-pairs that identify ranges to rehighlight (or
   * use GtkTextRegion, if possible) and merge them when possible */
  if(priv->is_user_action)
  {
    if(priv->mark_user_action == NULL)
    {
      priv->mark_user_action = gtk_text_buffer_create_mark(
        priv->buffer,
        NULL,
        &begin,
        TRUE
      );
    }
  }
  else
  {
    c4s_highlighter_rehighlight(C4S_HIGHLIGHTER(user_data), &begin, pos);
  }
}

static void
c4s_highlighter_erase_cb(G_GNUC_UNUSED GtkTextBuffer* buffer,
                         GtkTextIter* begin,
                         GtkTextIter* end,
                         gpointer user_data)
{
  /* begin == end */
  c4s_highlighter_rehighlight(C4S_HIGHLIGHTER(user_data), begin, end);
}

static void
c4s_highlighter_begin_user_action_cb(G_GNUC_UNUSED GtkTextBuffer* buffer,
                                     gpointer user_data)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;

  highlighter = C4S_HIGHLIGHTER(user_data);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  g_return_if_fail(priv->is_user_action == FALSE);

  priv->is_user_action = TRUE;
}

static void
c4s_highlighter_end_user_action_cb(GtkTextBuffer* buffer,
                                   gpointer user_data)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;
  GtkTextIter begin, end;
  
  highlighter = C4S_HIGHLIGHTER(user_data);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  g_return_if_fail(priv->is_user_action == TRUE);

  if(priv->mark_user_action != NULL)
  {
    gtk_text_buffer_get_iter_at_mark(buffer, &begin, priv->mark_user_action);

    gtk_text_buffer_get_iter_at_mark(
      buffer,
      &end,
      gtk_text_buffer_get_insert(buffer)
    );

    c4s_highlighter_rehighlight(highlighter, &begin, &end);
    gtk_text_buffer_delete_mark(buffer, priv->mark_user_action);
    priv->mark_user_action = NULL;
  }

  priv->is_user_action = FALSE;
}

static void
c4s_highlighter_reloaded_cb(G_GNUC_UNUSED C4sEngineFunctionList* list,
                            gpointer user_data)
{
  C4sHighlighter* highlighter;
  highlighter = user_data;
  
  c4s_highlighter_rehighlight_functions(highlighter);
}

static void
c4s_highlighter_notify_cb(GObject* object,
                          GParamSpec* pspec,
                          gpointer user_data)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;

  highlighter = C4S_HIGHLIGHTER(user_data);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  if(priv->buffer == NULL) return;

  if(strcmp(pspec->name, "color-comment") == 0)
  {
    g_object_set(
      priv->tag_single_comment,
      "foreground-gdk",
      c4s_preferences_get_color_comment(C4S_PREFERENCES(object)),
      NULL
    );

    g_object_set(
      priv->tag_multi_comment,
      "foreground-gdk",
      c4s_preferences_get_color_comment(C4S_PREFERENCES(object)),
      NULL
    );
  }
  else if(strcmp(pspec->name, "color-string") == 0)
  {
    g_object_set(
      priv->tag_string,
      "foreground-gdk",
      c4s_preferences_get_color_string(C4S_PREFERENCES(object)),
      NULL
    );
  }
  else if(strcmp(pspec->name, "color-keyword") == 0)
  {
    g_object_set(
      priv->tag_keyword,
      "foreground-gdk",
      c4s_preferences_get_color_keyword(C4S_PREFERENCES(object)),
      NULL
    );
  }
  else if(strcmp(pspec->name, "color-function") == 0)
  {
    g_object_set(
      priv->tag_function,
      "foreground-gdk",
      c4s_preferences_get_color_function(C4S_PREFERENCES(object)),
      NULL
    );
  }
  else if(strcmp(pspec->name, "color-number") == 0)
  {
    g_object_set(
      priv->tag_number,
      "foreground-gdk",
      c4s_preferences_get_color_number(C4S_PREFERENCES(object)),
      NULL
    );
  }
}

static void
c4s_highlighter_set_buffer(C4sHighlighter* highlighter,
                           GtkTextBuffer* buffer)
{
  C4sHighlighterPrivate* priv;
  GtkTextTagTable* table;
  GtkTextIter begin;
  GtkTextIter end;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  if(priv->buffer != NULL)
  {
    table = gtk_text_buffer_get_tag_table(priv->buffer);

    gtk_text_tag_table_remove(table, priv->tag_string);
    gtk_text_tag_table_remove(table, priv->tag_single_comment);
    gtk_text_tag_table_remove(table, priv->tag_multi_comment);
    gtk_text_tag_table_remove(table, priv->tag_keyword);
    gtk_text_tag_table_remove(table, priv->tag_function);
    gtk_text_tag_table_remove(table, priv->tag_number);
    gtk_text_tag_table_remove(table, priv->tag_comment);

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(table),
      G_CALLBACK(c4s_highlighter_insert_cb),
      highlighter
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(table),
      G_CALLBACK(c4s_highlighter_erase_cb),
      highlighter
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(table),
      G_CALLBACK(c4s_highlighter_begin_user_action_cb),
      highlighter
    );

    g_signal_handlers_disconnect_by_func(
      G_OBJECT(table),
      G_CALLBACK(c4s_highlighter_end_user_action_cb),
      highlighter
    );

    g_object_unref(priv->buffer);
  }

  priv->buffer = buffer;
  if(buffer == NULL) return;

  g_object_ref(buffer);

  priv->tag_string = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_single_comment = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_multi_comment = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_keyword = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_function = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_number = gtk_text_buffer_create_tag(buffer, NULL, NULL);
  priv->tag_comment = gtk_text_buffer_create_tag(buffer, NULL, NULL);

  g_signal_connect_after(
    G_OBJECT(buffer),
    "insert-text",
    G_CALLBACK(c4s_highlighter_insert_cb),
    highlighter
  );

  g_signal_connect_after(
    G_OBJECT(buffer),
    "delete-range",
    G_CALLBACK(c4s_highlighter_erase_cb),
    highlighter
  );

  g_signal_connect_after(
    G_OBJECT(buffer),
    "begin-user-action",
    G_CALLBACK(c4s_highlighter_begin_user_action_cb),
    highlighter
  );

  g_signal_connect_after(
    G_OBJECT(buffer),
    "end-user-action",
    G_CALLBACK(c4s_highlighter_end_user_action_cb),
    highlighter
  );

  c4s_highlighter_recolor(highlighter);

  gtk_text_buffer_get_start_iter(buffer, &begin);
  gtk_text_buffer_get_end_iter(buffer, &end);

  c4s_highlighter_rehighlight(highlighter, &begin, &end);
}

static void
c4s_highlighter_set_engine_function_list(C4sHighlighter* highlighter,
                                         C4sEngineFunctionList* list)
{
  C4sHighlighterPrivate* priv;
  GtkTextIter start;
  GtkTextIter end;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  
  if(priv->engine_function_list != NULL)
  {
    if(priv->buffer != NULL)
    {
      gtk_text_buffer_get_start_iter(priv->buffer, &start);
      gtk_text_buffer_get_end_iter(priv->buffer, &end);

      gtk_text_buffer_remove_tag(
        priv->buffer,
        priv->tag_function,
        &start,
        &end
      );
    }
    
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->engine_function_list),
      G_CALLBACK(c4s_highlighter_reloaded_cb),
      highlighter
    );

    g_object_unref(priv->engine_function_list);
  }
  
  priv->engine_function_list = list;
  if(list == NULL) return;
  
  g_object_ref(G_OBJECT(priv->engine_function_list));

  g_signal_connect(
    G_OBJECT(priv->engine_function_list),
    "reloaded",
    G_CALLBACK(c4s_highlighter_reloaded_cb),
    highlighter
  );
  
  c4s_highlighter_rehighlight_functions(highlighter);
}

static void
c4s_highlighter_init(C4sHighlighter* highlighter)
{
  C4sHighlighterPrivate* priv;
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  priv->buffer = NULL;
  priv->keywords = NULL;
  priv->preferences = NULL;

  priv->tag_string = NULL;
  priv->tag_single_comment = NULL;
  priv->tag_multi_comment = NULL;

  priv->tag_keyword = NULL;
  priv->tag_function = NULL;
  priv->tag_number = NULL;

  priv->tag_comment = NULL;

  priv->is_user_action = FALSE;
  priv->mark_user_action = NULL;
}

static void 
c4s_highlighter_finalize(GObject* object)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;

  highlighter = C4S_HIGHLIGHTER(object);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  /* cannot finalize this during a user action */
  g_assert(priv->mark_user_action == NULL);

  c4s_highlighter_set_buffer(highlighter, NULL);
  c4s_highlighter_set_engine_function_list(highlighter, NULL);

  g_signal_handlers_disconnect_by_func(
    G_OBJECT(priv->preferences),
    G_CALLBACK(c4s_highlighter_notify_cb),
    highlighter
  );

  g_object_unref(G_OBJECT(priv->preferences));

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_highlighter_set_property(GObject* object,
                             guint prop_id,
                             const GValue* value,
                             GParamSpec* pspec)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;
  GtkTextIter begin;
  GtkTextIter end;

  g_return_if_fail(C4S_IS_HIGHLIGHTER(object));

  highlighter = C4S_HIGHLIGHTER(object);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  switch(prop_id)
  {
  case PROP_BUFFER:
    c4s_highlighter_set_buffer(
      highlighter,
      GTK_TEXT_BUFFER(g_value_get_object(value))
    );

    break;
  case PROP_KEYWORDS:
    priv->keywords = (const gchar**)g_value_get_pointer(value);
    /* Rehighlight to highlight new keywords. */
    /* TODO: We also might just remove all keyword tags and rehighlight
     * only keywords */
    gtk_text_buffer_get_start_iter(priv->buffer, &begin);
    gtk_text_buffer_get_end_iter(priv->buffer, &end);

    c4s_highlighter_rehighlight(highlighter, &begin, &end);
    break;
  case PROP_ENGINE_FUNCTION_LIST:
    c4s_highlighter_set_engine_function_list(
      highlighter,
      C4S_ENGINE_FUNCTION_LIST(g_value_get_object(value))
    );

    break;
  case PROP_PREFERENCES:
    if(priv->preferences != NULL)
    {
      g_signal_handlers_disconnect_by_func(
        G_OBJECT(priv->preferences),
        G_CALLBACK(c4s_highlighter_notify_cb),
        highlighter
      );

      g_object_unref(priv->preferences);
    }

    priv->preferences = C4S_PREFERENCES(g_value_dup_object(value));

    g_signal_connect(
      G_OBJECT(priv->preferences),
      "notify",
      G_CALLBACK(c4s_highlighter_notify_cb),
      highlighter
    );

    c4s_highlighter_recolor(highlighter);

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_highlighter_get_property(GObject* object,
                             guint prop_id,
                             GValue* value,
                             GParamSpec* pspec)
{
  C4sHighlighter* highlighter;
  C4sHighlighterPrivate* priv;

  g_return_if_fail(C4S_IS_HIGHLIGHTER(object));

  highlighter = C4S_HIGHLIGHTER(object);
  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  switch(prop_id)
  {
  case PROP_BUFFER:
    g_value_set_object(value, priv->buffer);
    break;
  case PROP_KEYWORDS:
    g_value_set_pointer(value, priv->keywords);
    break;
  case PROP_ENGINE_FUNCTION_LIST:
    g_value_set_object(value, priv->engine_function_list);
    break;
  case PROP_PREFERENCES:
    g_value_set_object(value, priv->preferences);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_highlighter_class_init(C4sHighlighterClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sHighlighterPrivate));

  object_class->finalize = c4s_highlighter_finalize;
  object_class->set_property = c4s_highlighter_set_property;
  object_class->get_property = c4s_highlighter_get_property;

  klass->highlighted = c4s_highlighter_highlighted_cb;

  g_object_class_install_property(
    object_class,
    PROP_BUFFER,
    g_param_spec_object(
      "buffer",
      "Highlighted buffer",
      "GtkTextBuffer to apply highlighting tags to",
      GTK_TYPE_TEXT_BUFFER,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_KEYWORDS,
    g_param_spec_pointer(
      "keywords",
      "C4Script Keywords",
      "NULL-terminated keyword list",
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "Preferences from which to read color values",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_ENGINE_FUNCTION_LIST,
    g_param_spec_object(
      "engine-function-list",
      "C4Script function definitions",
      "C4Script function definitions to highlight engine functions",
      C4S_TYPE_ENGINE_FUNCTION_LIST,
      G_PARAM_READWRITE
    )
  );

  highlighter_signals[HIGHLIGHTED] = g_signal_new(
    "highlighted",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sHighlighterClass, highlighted),
    NULL, NULL,
    c4s_marshal_VOID__BOXED_BOXED,
    G_TYPE_NONE,
    2,
    GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE,
    GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE
  );
}

GType
c4s_highlighter_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sHighlighterClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_highlighter_class_init,
      NULL,
      NULL,
      sizeof(C4sHighlighter),
      0,
      (GInstanceInitFunc)c4s_highlighter_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sHighlighter",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sHighlighter*
c4s_highlighter_new(GtkTextBuffer* buffer,
                    const gchar** keywords,
                    C4sEngineFunctionList* engine_function_list,
                    C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_HIGHLIGHTER,
    "buffer", buffer,
    "keywords", keywords,
    "engine-function-list", engine_function_list,
    "preferences", preferences,
    NULL
  );

  return C4S_HIGHLIGHTER(object);
}

void
c4s_highlighter_forward_to_token(const C4sHighlighter* highlighter,
                                 GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  tag = c4s_highlighter_get_highlight_at(highlighter, iter);

  if(tag == priv->tag_single_comment)
    gtk_text_iter_forward_to_tag_toggle(iter, tag);
  else if(tag == priv->tag_multi_comment)
    gtk_text_iter_forward_to_tag_toggle(iter, tag);
  else if(tag == priv->tag_string) /* string is already a token */
    return;

  for(;;)
  {
    if(gtk_text_iter_begins_tag(iter, priv->tag_single_comment))
      gtk_text_iter_forward_to_tag_toggle(iter, priv->tag_single_comment);
    else if(gtk_text_iter_begins_tag(iter, priv->tag_multi_comment))
      gtk_text_iter_forward_to_tag_toggle(iter, priv->tag_multi_comment);
    else if(!c4s_forward_spaces(iter))
      break;
  }
}

void
c4s_highlighter_backward_to_token(const C4sHighlighter* highlighter,
                                  GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);

  tag = c4s_highlighter_get_highlight_at(highlighter, iter);

  if(tag == priv->tag_single_comment)
    gtk_text_iter_backward_to_tag_toggle(iter, tag);
  else if(tag == priv->tag_multi_comment)
    gtk_text_iter_backward_to_tag_toggle(iter, tag);
  else if(tag == priv->tag_string) /* string is already a token */
    return;

  for(;;)
  {
    if(gtk_text_iter_ends_tag(iter, priv->tag_single_comment))
      gtk_text_iter_backward_to_tag_toggle(iter, priv->tag_single_comment);
    else if(gtk_text_iter_ends_tag(iter, priv->tag_multi_comment))
      gtk_text_iter_backward_to_tag_toggle(iter, priv->tag_multi_comment);
    else if(!c4s_backward_spaces(iter))
      break;
  }
}

void
c4s_highlighter_forward_out_of_string(const C4sHighlighter* highlighter,
                                      GtkTextIter* iter)
{
  gtk_text_iter_forward_to_tag_toggle(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_string
  );
}

void
c4s_highlighter_backward_out_of_string(const C4sHighlighter* highlighter,
                                       GtkTextIter* iter)
{
  gtk_text_iter_backward_to_tag_toggle(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_string
  );
}

void
c4s_highlighter_forward_out_of_comment(const C4sHighlighter* highlighter,
                                       GtkTextIter* iter)
{
  gtk_text_iter_forward_to_tag_toggle(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_comment
  );
}

void
c4s_highlighter_backward_out_of_comment(const C4sHighlighter* highlighter,
                                        GtkTextIter* iter)
{
  gtk_text_iter_backward_to_tag_toggle(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_comment
  );
}

gboolean
c4s_highlighter_starts_string(const C4sHighlighter* highlighter,
                              const GtkTextIter* iter)
{
  return gtk_text_iter_begins_tag(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_string
  );
}

gboolean
c4s_highlighter_ends_string(const C4sHighlighter* highlighter,
                            const GtkTextIter* iter)
{
  return gtk_text_iter_ends_tag(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_string
  );
}

gboolean
c4s_highlighter_starts_comment(const C4sHighlighter* highlighter,
                               const GtkTextIter* iter)
{
  return gtk_text_iter_begins_tag(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_comment
  );
}

gboolean
c4s_highlighter_ends_comment(const C4sHighlighter* highlighter,
                             const GtkTextIter* iter)
{
  return gtk_text_iter_ends_tag(
    iter,
    C4S_HIGHLIGHTER_PRIVATE(highlighter)->tag_comment
  );
}

gboolean
c4s_highlighter_is_string(const C4sHighlighter* highlighter,
                          const GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  tag = c4s_highlighter_get_highlight_at(highlighter, iter);

  return tag == priv->tag_string;
}

gboolean
c4s_highlighter_is_comment(const C4sHighlighter* highlighter,
                           const GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  tag = c4s_highlighter_get_highlight_at(highlighter, iter);

  return (tag == priv->tag_single_comment) ||
         (tag == priv->tag_multi_comment);
}

gboolean
c4s_highlighter_is_code(const C4sHighlighter* highlighter,
                        const GtkTextIter* iter)
{
  C4sHighlighterPrivate* priv;
  GtkTextTag* tag;

  priv = C4S_HIGHLIGHTER_PRIVATE(highlighter);
  tag = c4s_highlighter_get_highlight_at(highlighter, iter);

  return tag == NULL;
}
