/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-iconmanager.h>
#include <c4scripter/icon/icon.h>

#include <gtk/gtkiconfactory.h>

typedef struct _C4sIconManagerPrivate C4sIconManagerPrivate;
struct _C4sIconManagerPrivate {
  GtkIconFactory* factory;

  GtkIconSet* set_logo;
  GtkIconSet* set_func_private;
  GtkIconSet* set_func_protected;
  GtkIconSet* set_func_public;
  GtkIconSet* set_func_global;
  GtkIconSet* set_documentation;
};

#define C4S_ICON_MANAGER_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_ICON_MANAGER, \
    C4sIconManagerPrivate \
  ))

enum {
  PROP_0,

  PROP_FACTORY
};

static GObjectClass* parent_class = NULL;

static void
c4s_icon_manager_load_icon(C4sIconManager* manager,
                           GtkIconSet** store,
                           const guint8 pixels[],
                           const gchar* stock_id)
{
  C4sIconManagerPrivate* priv;
  GdkPixbuf* pixbuf;
  
  priv = C4S_ICON_MANAGER_PRIVATE(manager);

  pixbuf = gdk_pixbuf_new_from_inline(-1, pixels, FALSE, NULL);
  *store = gtk_icon_set_new_from_pixbuf(pixbuf);
  g_object_unref(G_OBJECT(pixbuf));
  
  gtk_icon_factory_add(priv->factory, stock_id, *store);
}

static void
c4s_icon_manager_init(C4sIconManager* icon_manager)
{
  C4sIconManagerPrivate* priv;
  priv = C4S_ICON_MANAGER_PRIVATE(icon_manager);
  
  priv->factory = gtk_icon_factory_new();
  
  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_logo,
    icon_logo,
    C4S_STOCK_LOGO
  );

  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_func_private,
    icon_func_private,
    C4S_STOCK_FUNC_PRIVATE
  );
  
  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_func_protected,
    icon_func_protected,
    C4S_STOCK_FUNC_PROTECTED
  );
  
  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_func_public,
    icon_func_public,
    C4S_STOCK_FUNC_PUBLIC
  );
  
  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_func_global,
    icon_func_global,
    C4S_STOCK_FUNC_GLOBAL
  );
  
  c4s_icon_manager_load_icon(
    icon_manager,
    &priv->set_documentation,
    icon_documentation,
    C4S_STOCK_DOCUMENTATION
  );

  gtk_icon_factory_add_default(priv->factory);
}

static void 
c4s_icon_manager_finalize(GObject* object)
{
  C4sIconManager* icon_manager;
  C4sIconManagerPrivate* priv;

  icon_manager = C4S_ICON_MANAGER(object);
  priv = C4S_ICON_MANAGER_PRIVATE(icon_manager);

  gtk_icon_set_unref(priv->set_logo);
  gtk_icon_set_unref(priv->set_func_private);
  gtk_icon_set_unref(priv->set_func_protected);
  gtk_icon_set_unref(priv->set_func_public);
  gtk_icon_set_unref(priv->set_func_global);
  gtk_icon_set_unref(priv->set_documentation);
  
  g_object_unref(G_OBJECT(priv->factory));

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_icon_manager_set_property(GObject* object,
                          guint prop_id,
                          G_GNUC_UNUSED const GValue* value,
                          GParamSpec* pspec)
{
  C4sIconManager* icon_manager;
  C4sIconManagerPrivate* priv;

  g_return_if_fail(C4S_IS_ICON_MANAGER(object));

  icon_manager = C4S_ICON_MANAGER(object);
  priv = C4S_ICON_MANAGER_PRIVATE(icon_manager);

  switch(prop_id)
  {
    /* We have only read only properties */
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_icon_manager_get_property(GObject* object,
                              guint prop_id,
                              GValue* value,
                              GParamSpec* pspec)
{
  C4sIconManager* icon_manager;
  C4sIconManagerPrivate* priv;

  g_return_if_fail(C4S_IS_ICON_MANAGER(object));

  icon_manager = C4S_ICON_MANAGER(object);
  priv = C4S_ICON_MANAGER_PRIVATE(icon_manager);

  switch(prop_id)
  {
  case PROP_FACTORY:
    g_value_set_object(value, priv->factory);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_icon_manager_class_init(C4sIconManagerClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sIconManagerPrivate));

  object_class->finalize = c4s_icon_manager_finalize;
  object_class->set_property = c4s_icon_manager_set_property;
  object_class->get_property = c4s_icon_manager_get_property;

  g_object_class_install_property(
    object_class,
    PROP_FACTORY,
    g_param_spec_object(
      "icon-factory",
      "Icon factory",
      "The GtkIconFactory used by the icon manager",
      GTK_TYPE_ICON_FACTORY,
      G_PARAM_READABLE
    )
  );
}

GType
c4s_icon_manager_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sIconManagerClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_icon_manager_class_init,
      NULL,
      NULL,
      sizeof(C4sIconManager),
      0,
      (GInstanceInitFunc)c4s_icon_manager_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sIconManager",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sIconManager*
c4s_icon_manager_new(void)
{
  GObject* object;
  object = g_object_new(C4S_TYPE_ICON_MANAGER, NULL);
  return C4S_ICON_MANAGER(object);
}
