/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-preferences.h>

#include <pango/pango-font.h>

#include <stdio.h>

typedef struct _C4sPreferencesPrivate C4sPreferencesPrivate;
struct _C4sPreferencesPrivate {
  const C4sStorageBackendFuncs* backend_funcs;
  C4sStorageBackend* backend;

  guint tab_width;
  gboolean auto_indentation;
  gboolean highlight_brackets;

  GdkColor color_text;
  GdkColor color_string;
  GdkColor color_comment;
  GdkColor color_keyword;
  GdkColor color_function;
  GdkColor color_number;

  PangoFontDescription* font;

  gboolean view_tooltips;
  gboolean view_statusbar;

  C4sFunctionQueryAction function_query_action;
  gchar* chm_file_path;
};

#define C4S_PREFERENCES_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_PREFERENCES, \
    C4sPreferencesPrivate \
  ))

enum {
  PROP_0,

  PROP_TAB_WIDTH,
  PROP_AUTO_INDENTATION,
  PROP_HIGHLIGHT_BRACKETS,

  PROP_COLOR_TEXT,
  PROP_COLOR_STRING,
  PROP_COLOR_COMMENT,
  PROP_COLOR_KEYWORD,
  PROP_COLOR_FUNCTION,
  PROP_COLOR_NUMBER,

  PROP_FONT,

  PROP_VIEW_TOOLTIPS,
  PROP_VIEW_STATUSBAR,

  PROP_FUNCTION_QUERY_ACTION,
  PROP_CHM_FILE_PATH,

  /* construct */
  PROP_BACKEND
};

static GObjectClass* parent_class = NULL;

static gint
c4s_preferences_read_int(const C4sPreferences* preferences,
                         const gchar* group,
                         const gchar* key,
                         gint default_value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return default_value;

  return priv->backend_funcs->read_int(
    priv->backend,
    group,
    key,
    default_value
  );
}

static gboolean
c4s_preferences_read_boolean(const C4sPreferences* preferences,
                             const gchar* group,
                             const gchar* key,
                             gboolean default_value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return default_value;

  return priv->backend_funcs->read_boolean(
    priv->backend,
    group,
    key,
    default_value
  );
}

static gchar*
c4s_preferences_read_string(const C4sPreferences* preferences,
                            const gchar* group,
                            const gchar* key,
                            const gchar* default_value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return g_strdup(default_value);

  return priv->backend_funcs->read_string(
    priv->backend,
    group,
    key,
    default_value
  );
}

static void
c4s_preferences_read_color(const C4sPreferences* preferences,
                           const gchar* group,
                           const gchar* key,
                           const gchar* default_value,
                           GdkColor* color)
{
  C4sPreferencesPrivate* priv;
  gchar* spec;

  priv = C4S_PREFERENCES_PRIVATE(preferences);
  if(priv->backend == NULL)
  {
    gdk_color_parse(default_value, color);
    return;
  }

  spec = priv->backend_funcs->read_string(
    priv->backend,
    group,
    key,
    default_value
  );

  gdk_color_parse(spec, color);
  g_free(spec);
}

static void
c4s_preferences_write_int(const C4sPreferences* preferences,
                          const gchar* group,
                          const gchar* key,
                          gint value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return;

  priv->backend_funcs->write_int(
    priv->backend,
    group,
    key,
    value
  );
}

static void
c4s_preferences_write_boolean(const C4sPreferences* preferences,
                              const gchar* group,
                              const gchar* key,
                              gboolean value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return;

  priv->backend_funcs->write_boolean(
    priv->backend,
    group,
    key,
    value
  );
}

static void
c4s_preferences_write_string(const C4sPreferences* preferences,
                             const gchar* group,
                             const gchar* key,
                             const gchar* value)
{
  C4sPreferencesPrivate* priv;
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  if(priv->backend == NULL) return;

  priv->backend_funcs->write_string(
    priv->backend,
    group,
    key,
    value
  );
}

static void
c4s_preferences_write_color(const C4sPreferences* preferences,
                            const gchar* group,
                            const gchar* key,
                            const GdkColor* value)
{
  gchar buffer[8];
  C4sPreferencesPrivate* priv;

  priv = C4S_PREFERENCES_PRIVATE(preferences);
  if(priv->backend == NULL) return;

  sprintf(
    buffer,
    "#%.2x%.2x%.2x",
    value->red * 0xff / 0xffff,
    value->green * 0xff / 0xffff,
    value->blue * 0xff / 0xffff
  );

  priv->backend_funcs->write_string(
    priv->backend,
    group,
    key,
    buffer
  );
}

static GObject*
c4s_preferences_constructor(GType type,
                            guint n_construct_properties,
                            GObjectConstructParam *construct_param)
{
  GObject* g_object;
  C4sPreferences* preferences;
  C4sPreferencesPrivate* priv;
  GError* error;
  gchar* font;

  g_object = G_OBJECT_CLASS(parent_class)->constructor(
    type,
    n_construct_properties,
    construct_param
  );

  /* Now instance_init has been called, so priv has been initialized. The
   * construct properties also have been set, so we can initialize the
   * backend */
  preferences = C4S_PREFERENCES(g_object);
  priv = C4S_PREFERENCES_PRIVATE(preferences);
  error = NULL;

  priv->backend = priv->backend_funcs->init(&error);
  if(priv->backend == NULL)
  {
    g_warning(
      "Failed to initialize preference's storage backend: %s\n",
      error->message
    );

    g_error_free(error);
  }

  priv->tab_width = c4s_preferences_read_int(
    preferences,
    "settings",
    "tab-width",
    2
  );

  priv->auto_indentation = c4s_preferences_read_boolean(
    preferences,
    "settings",
    "auto-indentation",
    TRUE
  );

  priv->highlight_brackets = c4s_preferences_read_boolean(
    preferences,
    "settings",
    "highlight-bracket",
    TRUE
  );

  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-text",
    "#000000",
    &priv->color_text
  );

  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-string",
    "#C00000",
    &priv->color_string
  );

  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-comment",
    "#008000",
    &priv->color_comment
  );

  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-keyword",
    "#0000FF",
    &priv->color_keyword
  );
  
  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-function",
    "#0080C0",
    &priv->color_function
  );

  c4s_preferences_read_color(
    preferences,
    "settings",
    "color-number",
    "#C00080",
    &priv->color_number
  );

  font = c4s_preferences_read_string(
    preferences,
    "settings",
    "font",
    "Monospace 10"
  );

  priv->view_tooltips = c4s_preferences_read_boolean(
    preferences,
    "view",
    "tooltips",
    TRUE
  );

  priv->view_statusbar = c4s_preferences_read_boolean(
    preferences,
    "view",
    "statusbar",
    TRUE
  );
  
  priv->function_query_action = c4s_preferences_read_int(
    preferences,
    "functions",
    "query-action",
    C4S_FUNCTION_QUERY_ACTION_ONLINE
  );

  priv->chm_file_path = c4s_preferences_read_string(
    preferences,
    "functions",
    "chm-file",
    ""
  );

  priv->font = pango_font_description_from_string(font);
  g_free(font);

  return g_object;
}

static void 
c4s_preferences_finalize(GObject* object)
{
  C4sPreferences* preferences;
  C4sPreferencesPrivate* priv;

  preferences = C4S_PREFERENCES(object);
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  priv->backend_funcs->deinit(priv->backend);

  pango_font_description_free(priv->font);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_preferences_set_property(GObject* object,
                             guint prop_id,
                             const GValue* value,
                             GParamSpec* pspec)
{
  C4sPreferences* preferences;
  C4sPreferencesPrivate* priv;
  gpointer* pointer;

  g_return_if_fail(C4S_IS_PREFERENCES(object));

  preferences = C4S_PREFERENCES(object);
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  switch(prop_id)
  {
  case PROP_TAB_WIDTH:
    priv->tab_width = g_value_get_uint(value);
    break;
  case PROP_AUTO_INDENTATION:
    priv->auto_indentation = g_value_get_boolean(value);
    break;
  case PROP_HIGHLIGHT_BRACKETS:
    priv->highlight_brackets = g_value_get_boolean(value);
    break;
  case PROP_COLOR_TEXT:
    pointer = g_value_get_boxed(value);
    priv->color_text = *((GdkColor*)pointer);
    break;
  case PROP_COLOR_STRING:
    pointer = g_value_get_boxed(value);
    priv->color_string = *((GdkColor*)pointer);
    break;
  case PROP_COLOR_COMMENT:
    pointer = g_value_get_boxed(value);
    priv->color_comment = *((GdkColor*)pointer);
    break;
  case PROP_COLOR_KEYWORD:
    pointer = g_value_get_boxed(value);
    priv->color_keyword = *((GdkColor*)pointer);
    break;
  case PROP_COLOR_FUNCTION:
    pointer = g_value_get_boxed(value);
    priv->color_function = *((GdkColor*)pointer);
    break;
  case PROP_COLOR_NUMBER:
    pointer = g_value_get_boxed(value);
    priv->color_number = *((GdkColor*)pointer);
    break;
  case PROP_FONT:
    pointer = g_value_get_boxed(value);
    pango_font_description_free(priv->font);
    priv->font = pango_font_description_copy(
      (PangoFontDescription*)pointer
    );
    break;
  case PROP_VIEW_TOOLTIPS:
    priv->view_tooltips = g_value_get_boolean(value);
    break;
  case PROP_VIEW_STATUSBAR:
    priv->view_statusbar = g_value_get_boolean(value);
    break;
  case PROP_FUNCTION_QUERY_ACTION:
    priv->function_query_action = g_value_get_enum(value);
    break;
  case PROP_CHM_FILE_PATH:
    g_free(priv->chm_file_path);
    priv->chm_file_path = g_value_dup_string(value);
    break;
  case PROP_BACKEND:
    priv->backend_funcs = g_value_get_pointer(value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_get_property(GObject* object,
                             guint prop_id,
                             GValue* value,
                             GParamSpec* pspec)
{
  C4sPreferences* preferences;
  C4sPreferencesPrivate* priv;

  g_return_if_fail(C4S_IS_PREFERENCES(object));

  preferences = C4S_PREFERENCES(object);
  priv = C4S_PREFERENCES_PRIVATE(preferences);

  switch(prop_id)
  {
  case PROP_TAB_WIDTH:
    g_value_set_uint(value, priv->tab_width);
    break;
  case PROP_AUTO_INDENTATION:
    g_value_set_boolean(value, priv->auto_indentation);
    break;
  case PROP_HIGHLIGHT_BRACKETS:
    g_value_set_boolean(value, priv->highlight_brackets);
    break;
  case PROP_COLOR_TEXT:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_text)
    );
    break;
  case PROP_COLOR_STRING:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_string)
    );
    break;
  case PROP_COLOR_COMMENT:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_comment)
    );
    break;
  case PROP_COLOR_KEYWORD:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_keyword)
    );
    break;
  case PROP_COLOR_FUNCTION:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_function)
    );
    break;
  case PROP_COLOR_NUMBER:
    g_value_set_boxed(
      value,
      gdk_color_copy(&priv->color_number)
    );
    break;
  case PROP_FONT:
    g_value_set_boxed(
      value,
      pango_font_description_copy(priv->font)
    );
    break;
  case PROP_VIEW_TOOLTIPS:
    g_value_set_boolean(value, priv->view_tooltips);
    break;
  case PROP_VIEW_STATUSBAR:
    g_value_set_boolean(value, priv->view_statusbar);
    break;
  case PROP_FUNCTION_QUERY_ACTION:
    g_value_set_enum(value, priv->function_query_action);
    break;
  case PROP_CHM_FILE_PATH:
    g_value_set_string(value, priv->chm_file_path);
    break;
  case PROP_BACKEND:
    g_value_set_pointer(value, (gpointer*)priv->backend_funcs);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_class_init(C4sPreferencesClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sPreferencesPrivate));

  object_class->constructor = c4s_preferences_constructor;
  object_class->finalize = c4s_preferences_finalize;
  object_class->set_property = c4s_preferences_set_property;
  object_class->get_property = c4s_preferences_get_property;

  g_object_class_install_property(
    object_class,
    PROP_TAB_WIDTH,
    g_param_spec_uint(
      "tab-width",
      "Tab width",
      "Tabulator width",
      1,
      8,
      2,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_AUTO_INDENTATION,
    g_param_spec_boolean(
      "auto-indentation",
      "Auto indentation",
      "Automatic indentation",
      TRUE,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_HIGHLIGHT_BRACKETS,
    g_param_spec_boolean(
      "highlight-brackets",
      "Highlight brackets",
      "Highlight matching brackets",
      TRUE,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_COLOR_TEXT,
    g_param_spec_boxed(
      "color-text",
      "Text color",
      "Color for normal text",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_COLOR_STRING,
    g_param_spec_boxed(
      "color-string",
      "String color",
      "Color used for highlighting string literals",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_COLOR_COMMENT,
    g_param_spec_boxed(
      "color-comment",
      "Comment color",
      "Color used for highlighting comments",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_COLOR_KEYWORD,
    g_param_spec_boxed(
      "color-keyword",
      "Keyword color",
      "Color used for highlighting keywords",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_COLOR_FUNCTION,
    g_param_spec_boxed(
      "color-function",
      "Function color",
      "Color used for highilghting functions",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_COLOR_NUMBER,
    g_param_spec_boxed(
      "color-number",
      "Number color",
      "Color used for highlighting numbers",
      GDK_TYPE_COLOR,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_FONT,
    g_param_spec_boxed(
      "font",
      "Font",
      "Font to display text with",
      PANGO_TYPE_FONT_DESCRIPTION,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_VIEW_TOOLTIPS,
    g_param_spec_boolean(
      "view-tooltips",
      "Show tooltips",
      "Whether to show parameter tooltips",
      TRUE,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_VIEW_STATUSBAR,
    g_param_spec_boolean(
      "view-statusbar",
      "Show statusbar",
      "Whether to show the statusbar",
      TRUE,
      G_PARAM_READWRITE
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_FUNCTION_QUERY_ACTION,
    g_param_spec_enum(
      "function-query-action",
      "Function query action",
      "Mode that is used to query C4Script function definitions",
      C4S_TYPE_FUNCTION_QUERY_ACTION,
      C4S_FUNCTION_QUERY_ACTION_ONLINE,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_CHM_FILE_PATH,
    g_param_spec_string(
      "chm-file-path",
      "CHM file path",
      "Location to Clonk developer documentation CHM file",
      "",
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_BACKEND,
    g_param_spec_pointer(
      "backend",
      "Backend",
      "Storage backend to use",
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );
}

GType
c4s_preferences_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sPreferencesClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_preferences_class_init,
      NULL,
      NULL,
      sizeof(C4sPreferences),
      0,
      NULL /*(GInstanceInitFunc)c4s_preferences_init*/,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sPreferences",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sPreferences*
c4s_preferences_new(const C4sStorageBackendFuncs* backend)
{
  C4sPreferences* preferences;
  preferences = g_object_new(C4S_TYPE_PREFERENCES, "backend", backend, NULL);
  return preferences;
}

void
c4s_preferences_save(const C4sPreferences* preferences)
{
  C4sPreferencesPrivate* priv;
  gchar* font;

  priv = C4S_PREFERENCES_PRIVATE(preferences);
  font = pango_font_description_to_string(priv->font);

  c4s_preferences_write_int(
    preferences,
    "settings",
    "tab-width",
    priv->tab_width
  );

  c4s_preferences_write_boolean(
    preferences,
    "settings",
    "auto-indentation",
    priv->auto_indentation
  );

  c4s_preferences_write_boolean(
    preferences,
    "settings",
    "highlight-brackets",
    priv->highlight_brackets
  );

  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-text",
    &priv->color_text
  );

  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-string",
    &priv->color_string
  );

  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-comment",
    &priv->color_comment
  );

  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-keyword",
    &priv->color_keyword
  );
  
  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-function",
    &priv->color_function
  );

  c4s_preferences_write_color(
    preferences,
    "settings",
    "color-number",
    &priv->color_number
  );

  c4s_preferences_write_string(
    preferences,
    "settings",
    "font",
    font
  );

  c4s_preferences_write_boolean(
    preferences,
    "view",
    "tooltips",
    priv->view_tooltips
  );

  c4s_preferences_write_boolean(
    preferences,
    "view",
    "statusbar",
    priv->view_statusbar
  );
  
  c4s_preferences_write_int(
    preferences,
    "functions",
    "query-action",
    priv->function_query_action
  );

  c4s_preferences_write_string(
    preferences,
    "functions",
    "chm-file",
    priv->chm_file_path
  );

  g_free(font);
}

guint
c4s_preferences_get_tab_width(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->tab_width;
}

gboolean
c4s_preferences_get_auto_indentation(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->auto_indentation;
}

gboolean
c4s_preferences_get_highlight_brackets(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->highlight_brackets;
}

const GdkColor*
c4s_preferences_get_color_text(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_text;
}

const GdkColor*
c4s_preferences_get_color_string(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_string;
}

const GdkColor*
c4s_preferences_get_color_comment(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_comment;
}

const GdkColor*
c4s_preferences_get_color_keyword(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_keyword;
}

const GdkColor*
c4s_preferences_get_color_function(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_function;
}

const GdkColor*
c4s_preferences_get_color_number(const C4sPreferences* preferences)
{
  return &C4S_PREFERENCES_PRIVATE(preferences)->color_number;
}

const PangoFontDescription*
c4s_preferences_get_font(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->font;
}

gboolean
c4s_preferences_get_view_tooltips(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->view_tooltips;
}

gboolean
c4s_preferences_get_view_statusbar(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->view_statusbar;
}

C4sFunctionQueryAction
c4s_preferences_get_function_query_action(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->function_query_action;
}

const gchar*
c4s_preferences_get_chm_file_path(const C4sPreferences* preferences)
{
  return C4S_PREFERENCES_PRIVATE(preferences)->chm_file_path;
}

void
c4s_preferences_set_tab_width(C4sPreferences* preferences,
                              guint tab_width)
{
  C4S_PREFERENCES_PRIVATE(preferences)->tab_width = tab_width;
  g_object_notify(G_OBJECT(preferences), "tab-width");
}

void
c4s_preferences_set_auto_indentation(C4sPreferences* preferences,
                                     gboolean auto_indentation)
{
  C4S_PREFERENCES_PRIVATE(preferences)->auto_indentation = auto_indentation;
  g_object_notify(G_OBJECT(preferences), "auto-indentation");
}

void
c4s_preferences_set_highlight_brackets(C4sPreferences* preferences,
                                       gboolean highlight_brackets)
{
  C4S_PREFERENCES_PRIVATE(preferences)->highlight_brackets =
    highlight_brackets;

  g_object_notify(G_OBJECT(preferences), "highlight-brackets");
}

void
c4s_preferences_set_color_text(C4sPreferences* preferences,
                               const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_text = *color;
  g_object_notify(G_OBJECT(preferences), "color-text");
}

void
c4s_preferences_set_color_string(C4sPreferences* preferences,
                                 const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_string = *color;
  g_object_notify(G_OBJECT(preferences), "color-string");
}

void
c4s_preferences_set_color_comment(C4sPreferences* preferences,
                                  const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_comment = *color;
  g_object_notify(G_OBJECT(preferences), "color-comment");
}

void
c4s_preferences_set_color_keyword(C4sPreferences* preferences,
                                  const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_keyword = *color;
  g_object_notify(G_OBJECT(preferences), "color-keyword");
}

void
c4s_preferences_set_color_function(C4sPreferences* preferences,
                                   const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_function = *color;
  g_object_notify(G_OBJECT(preferences), "color-function");
}

void
c4s_preferences_set_color_number(C4sPreferences* preferences,
                                 const GdkColor* color)
{
  C4S_PREFERENCES_PRIVATE(preferences)->color_number = *color;
  g_object_notify(G_OBJECT(preferences), "color-number");
}

void
c4s_preferences_set_font(C4sPreferences* preferences,
                         const PangoFontDescription* font)
{
  pango_font_description_free(C4S_PREFERENCES_PRIVATE(preferences)->font);
  C4S_PREFERENCES_PRIVATE(preferences)->font =
    pango_font_description_copy(font);

  g_object_notify(G_OBJECT(preferences), "font");
}

void
c4s_preferences_set_view_tooltips(C4sPreferences* preferences,
                                  gboolean view_tooltips)
{
  C4S_PREFERENCES_PRIVATE(preferences)->view_tooltips = view_tooltips;
  g_object_notify(G_OBJECT(preferences), "view-tooltips");
}

void
c4s_preferences_set_view_statusbar(C4sPreferences* preferences,
                                   gboolean view_statusbar)
{
  C4S_PREFERENCES_PRIVATE(preferences)->view_statusbar = view_statusbar;
  g_object_notify(G_OBJECT(preferences), "view-statusbar");
}

void
c4s_preferences_set_function_query_action(C4sPreferences* preferences,
                                          C4sFunctionQueryAction action)
{
  C4S_PREFERENCES_PRIVATE(preferences)->function_query_action = action;
  g_object_notify(G_OBJECT(preferences), "function-query-action");
}

void
c4s_preferences_set_chm_file_path(C4sPreferences* preferences,
                                  const gchar* file_path)
{
  g_free(C4S_PREFERENCES_PRIVATE(preferences)->chm_file_path);
  C4S_PREFERENCES_PRIVATE(preferences)->chm_file_path = g_strdup(file_path);
  g_object_notify(G_OBJECT(preferences), "chm-file-path");
}
