/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-remotecomm.h>
#include <c4scripter/c4s-marshal.h>

struct _C4sRemoteComm {
  GObject parent_instance;

  const C4sCommBackendFuncs* backend_funcs;
  C4sCommBackend* backend;
};

enum {
  FILE_RECEIVED,
  LAST_SIGNAL
};

static GObjectClass* parent_class = NULL;
static guint remote_comm_signals[LAST_SIGNAL];

static void
c4s_remote_comm_init(C4sRemoteComm* comm)
{
  comm->backend_funcs = NULL;
  comm->backend = NULL;
}

static void 
c4s_remote_comm_finalize(GObject* object)
{
  C4sRemoteComm* comm;
  comm = C4S_REMOTE_COMM(object);

  if(comm->backend != NULL)
    comm->backend_funcs->deinit(comm->backend);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_remote_comm_file_received(G_GNUC_UNUSED C4sRemoteComm* comm,
                              G_GNUC_UNUSED const gchar* file)
{
  /* Default handler has nothing to do */
}

static void
c4s_remote_comm_class_init(C4sRemoteCommClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  object_class->finalize = c4s_remote_comm_finalize;
  
  klass->file_received = c4s_remote_comm_file_received;
  
  remote_comm_signals[FILE_RECEIVED] = g_signal_new(
    "file-received",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sRemoteCommClass, file_received),
    NULL, NULL,
    c4s_marshal_VOID__STRING,
    G_TYPE_NONE,
    1,
    G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE
  );
}

static void
c4s_remote_comm_file_func(const gchar* file,
                          gpointer data)
{
  C4sRemoteComm* comm;
  gchar* conv_file;

  comm = C4S_REMOTE_COMM(data);

  /* TODO: Error handling */
  conv_file = g_filename_from_utf8(file, -1, NULL, NULL, NULL);

  g_signal_emit(
    G_OBJECT(comm),
    remote_comm_signals[FILE_RECEIVED],
    0,
    conv_file
  );

  g_free(conv_file);
}

GType
c4s_remote_comm_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sRemoteCommClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_remote_comm_class_init,
      NULL,
      NULL,
      sizeof(C4sRemoteComm),
      0,
      (GInstanceInitFunc)c4s_remote_comm_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sRemoteComm",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sRemoteComm*
c4s_remote_comm_new(const C4sCommBackendFuncs* backend)
{
  C4sRemoteComm* remote_comm;
  remote_comm = g_object_new(C4S_TYPE_REMOTE_COMM, NULL);

  remote_comm->backend_funcs = backend;
  remote_comm->backend = remote_comm->backend_funcs->init(
    c4s_remote_comm_file_func,
    remote_comm
  );
  
  return remote_comm;
}

gboolean
c4s_remote_comm_connect(C4sRemoteComm* comm)
{
  return comm->backend_funcs->connect(comm->backend);
}

gboolean
c4s_remote_comm_open(C4sRemoteComm* comm,
                     GtkWindow* window)
{
  return comm->backend_funcs->open(comm->backend, window);
}

void
c4s_remote_comm_send(C4sRemoteComm* comm,
                     int argc,
                     gchar* const *argv)
{
  gchar** abs_paths;
  gchar* cur_dir;
  gchar* path;
  int i;

  cur_dir = g_get_current_dir();
  abs_paths = g_new(gchar*, argc);
  
  for(i = 0; i < argc; ++ i)
  {
    if(g_path_is_absolute(argv[i]))
      path = g_strdup(argv[i]);
    else
      path = g_build_filename(cur_dir, argv[i], NULL);

    /* TODO: What encoding is this filename in? */
    /* TODO: Error handling */
    abs_paths[i] = g_locale_to_utf8(path, -1, NULL, NULL, NULL);
    g_free(path);
  }
  
  g_free(cur_dir);
  comm->backend_funcs->send(comm->backend, argc, abs_paths);

  for(i = 0; i < argc; ++ i)
    g_free(abs_paths[i]);
  g_free(abs_paths);
}
