/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-tcp.h>

#ifdef _WIN32
# include <winsock2.h>
#else
# include <sys/socket.h>
# include <sys/types.h>
# include <netdb.h>
# include <unistd.h>
# include <signal.h>
# include <errno.h>
# include <string.h>
#endif

typedef enum _C4sTcpConnectorStatus {
  C4S_TCP_CONNECTOR_RUNNING,
  C4S_TCP_CONNECTOR_EXITED
} C4sTcpConnectorStatus;

typedef struct _C4sTcpConnectorData C4sTcpConnectorData;
struct _C4sTcpConnectorData {
  GThread* worker;

  C4sTcpConnector* connector;
  C4sTcpConnectorStatus status;        
  gchar* hostname;
  guint port;

  GError* error;
  GIOChannel* channel;
};

struct _C4sTcpConnector {
  C4sTcpConnectorData* data;

  C4sTcpResolvFunc resolv_func;
  C4sTcpConnectFunc connect_func;
  C4sTcpErrorFunc error_func;

  gpointer user_data;
};

typedef struct _C4sTcpConnectionPrivate C4sTcpConnectionPrivate;
struct _C4sTcpConnectionPrivate {
  GIOChannel* channel;
  guint io_handle;

  C4sTcpIOFunc io_func;
  gpointer user_data;
};

#define C4S_TCP_CONNECTION_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_TCP_CONNECTION, \
    C4sTcpConnectionPrivate \
  ))

enum {
  PROP_0,

  /* construct only */
  PROP_IO_CHANNEL
};

#ifdef _WIN32
typedef SOCKET C4sTcpNativeSocket;
#else
typedef int C4sTcpNativeSocket;
static const C4sTcpNativeSocket INVALID_SOCKET = -1;
#endif

static GObjectClass* parent_class = NULL;

static void
c4s_tcp_connector_set_error(GError** error,
                            gboolean from_gethostbyname)
{
#ifdef _WIN32
  LPVOID msgbuf;
  gchar* converted;
  gint errcode;

  errcode = WSAGetLastError();

  FormatMessageA(
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    errcode,
    MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
    (LPSTR)&msgbuf,
    0,
    NULL
  );

  g_set_error(
    error,
    g_quark_from_static_string("C4S_TCP_CONNECTION"),
    errcode,
    "%s",
    msgbuf
  );

  LocalFree(msgbuf);
#else
  gint errcode;

  if(from_gethostbyname == TRUE)
  {
    errcode = h_errno;

    g_set_error(
      error,
      g_quark_from_static_string("C4S_TCP_CONNECTION"),
      errcode,
      "%s",
      hstrerror(errcode)
    );
  }
  else
  {
    errcode = errno;

    g_set_error(
      error,
      g_quark_from_static_string("C4S_TCP_CONNECTION"),
      errcode,
      "%s",
      strerror(errno)
    );
  }
#endif
}

static void
c4s_tcp_connector_dispatch_clear(C4sTcpConnectorData* data)
{
  g_thread_join(data->worker);

  /* free connector if still running */
  if(data->status == C4S_TCP_CONNECTOR_RUNNING)
    c4s_tcp_cancel(data->connector);

  g_free(data->hostname);
  g_free(data);
}

static gboolean
c4s_tcp_connector_dispatch_resolv(gpointer user_data)
{
  C4sTcpConnectorData* data;
  data = user_data;

  if(data->status == C4S_TCP_CONNECTOR_RUNNING)
    data->connector->resolv_func(data->connector->user_data);

  return FALSE;
}

static gboolean
c4s_tcp_connector_dispatch_connect(gpointer user_data)
{
  C4sTcpConnectorData* data;
  C4sTcpConnection* conn;
  GIOChannel* channel;

  data = user_data;
  channel = data->channel;

  if(data->status == C4S_TCP_CONNECTOR_RUNNING)
  {
    conn = g_object_new(C4S_TYPE_TCP_CONNECTION, "io-channel", channel, NULL);
    data->connector->connect_func(conn, data->connector->user_data);
    g_object_unref(conn);
  }

  g_io_channel_unref(channel);
  c4s_tcp_connector_dispatch_clear(data);
  return FALSE;
}

static gboolean
c4s_tcp_connector_dispatch_error(gpointer user_data)
{
  C4sTcpConnectorData* data;

  data = user_data;
  if(data->status == C4S_TCP_CONNECTOR_RUNNING)
    data->connector->error_func(data->error, data->connector->user_data);

  c4s_tcp_connector_dispatch_clear(data);
  return FALSE;
}

static void
c4s_tcp_close(C4sTcpNativeSocket sock)
{
#ifdef _WIN32
  closesocket(sock);
#else
  close(sock);
#endif
}

static gpointer
c4s_tcp_connector_thread_func(gpointer user_data)
{
  C4sTcpConnectorData* data;
  C4sTcpNativeSocket sock;
  struct sockaddr_in addr;
  struct hostent* ent;

  data = user_data;
  ent = gethostbyname(data->hostname);
  if(ent == NULL)
  {
    c4s_tcp_connector_set_error(&data->error, TRUE);
    g_idle_add(c4s_tcp_connector_dispatch_error, data);
    return NULL;
  }

  g_idle_add(c4s_tcp_connector_dispatch_resolv, data);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(data->port);
  memcpy(&addr.sin_addr, ent->h_addr_list[0], sizeof(struct in_addr));

  sock = socket(PF_INET, SOCK_STREAM, 0);
  if(sock == INVALID_SOCKET)
  {
    c4s_tcp_connector_set_error(&data->error, FALSE);
    g_idle_add(c4s_tcp_connector_dispatch_error, data);
    return NULL;
  }

  if(connect(sock, (struct sockaddr*)&addr, sizeof(addr)) == -1)
  {
    c4s_tcp_connector_set_error(&data->error, FALSE);
    g_idle_add(c4s_tcp_connector_dispatch_error, data);
    c4s_tcp_close(sock);
    return NULL;
  }

#ifdef _WIN32
  data->channel = g_io_channel_win32_new_socket(sock);
#else
  data->channel = g_io_channel_unix_new(sock);
#endif

  g_io_channel_set_close_on_unref(data->channel, TRUE);
  g_io_channel_set_encoding(data->channel, NULL, NULL);
  g_io_channel_set_buffered(data->channel, FALSE);
  g_io_channel_set_flags(
    data->channel,
    g_io_channel_get_flags(data->channel) | G_IO_FLAG_NONBLOCK,
    NULL
  );

  g_idle_add(c4s_tcp_connector_dispatch_connect, data);
  return NULL;
}

static gboolean
c4s_tcp_connection_io_func(GIOChannel* source,
                           GIOCondition condition,
                           gpointer user_data)
{
  C4sTcpConnection* conn;
  C4sTcpConnectionPrivate* priv;

  conn = C4S_TCP_CONNECTION(user_data);
  priv = C4S_TCP_CONNECTION_PRIVATE(conn);

  priv->io_func(conn, condition, priv->user_data);
  return TRUE;
}

#ifdef _WIN32
static gboolean
c4s_tcp_connection_noop_func(gpointer user_data)
{
  return FALSE;
}
#endif

static void
c4s_tcp_connection_init(C4sTcpConnection* tcp_connection)
{
  C4sTcpConnectionPrivate* priv;
  priv = C4S_TCP_CONNECTION_PRIVATE(tcp_connection);

  priv->channel = NULL;
  priv->io_handle = 0;
}

static GObject*
c4s_tcp_connection_constructor(GType type,
                               guint n_construct_properties,
                               GObjectConstructParam *construct_param)
{
  GObject* g_object;
  C4sTcpConnection* tcp_connection;
  C4sTcpConnectionPrivate* priv;

  g_object = G_OBJECT_CLASS(parent_class)->constructor(
    type,
    n_construct_properties,
    construct_param
  );

  tcp_connection = C4S_TCP_CONNECTION(g_object);
  priv = C4S_TCP_CONNECTION_PRIVATE(tcp_connection);

  /* Must have been set as construct only property */
  g_assert(priv->channel != NULL);

  /* Is already refed in set_property */
  /*g_io_channel_ref(priv->channel);*/
  return g_object;
}

static void 
c4s_tcp_connection_finalize(GObject* object)
{
  C4sTcpConnection* tcp_connection;
  C4sTcpConnectionPrivate* priv;

  tcp_connection = C4S_TCP_CONNECTION(object);
  priv = C4S_TCP_CONNECTION_PRIVATE(tcp_connection);

  if(priv->io_handle != 0)
    g_source_remove(priv->io_handle);

  g_io_channel_unref(priv->channel);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_tcp_connection_set_property(GObject* object,
                                guint prop_id,
                                const GValue* value,
                                GParamSpec* pspec)
{
  C4sTcpConnection* tcp_connection;
  C4sTcpConnectionPrivate* priv;

  g_return_if_fail(C4S_IS_TCP_CONNECTION(object));

  tcp_connection = C4S_TCP_CONNECTION(object);
  priv = C4S_TCP_CONNECTION_PRIVATE(tcp_connection);

  switch(prop_id)
  {
  case PROP_IO_CHANNEL:
    if(priv->channel != NULL)
      g_io_channel_unref(priv->channel);

    priv->channel = g_value_get_pointer(value);
    g_io_channel_ref(priv->channel);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_tcp_connection_get_property(GObject* object,
                                guint prop_id,
                                GValue* value,
                                GParamSpec* pspec)
{
  C4sTcpConnection* tcp_connection;
  C4sTcpConnectionPrivate* priv;

  g_return_if_fail(C4S_IS_TCP_CONNECTION(object));

  tcp_connection = C4S_TCP_CONNECTION(object);
  priv = C4S_TCP_CONNECTION_PRIVATE(tcp_connection);

  switch(prop_id)
  {
  case PROP_IO_CHANNEL:
    /* TODO: Are we supposed to ref channel here? */
    g_value_set_pointer(value, priv->channel);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_tcp_connection_class_init(C4sTcpConnectionClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sTcpConnectionPrivate));

  object_class->constructor = c4s_tcp_connection_constructor;
  object_class->finalize = c4s_tcp_connection_finalize;
  object_class->set_property = c4s_tcp_connection_set_property;
  object_class->get_property = c4s_tcp_connection_get_property;

  g_object_class_install_property(
    object_class,
    PROP_IO_CHANNEL,
    g_param_spec_pointer(
      "io-channel",
      "IO channel",
      "IO channel used for communication",
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );
}

gboolean
c4s_tcp_init(GError** error)
{
#ifdef _WIN32
  WSADATA data;
  WORD wVersionRequested;
  int err_code;

  wVersionRequested = MAKEWORD(2, 2);
  err_code = WSAStartup(wVersionRequested, &data);

  if(err_code != 0)
  {
    c4s_tcp_connector_set_error(error, FALSE);
    return FALSE;
  }

  return TRUE;
#else
  signal(SIGPIPE, SIG_IGN);
  return TRUE;
#endif
}

void
c4s_tcp_deinit(void)
{
#ifdef _WIN32
  WSACleanup();
#endif
}

C4sTcpConnector*
c4s_tcp_connect(const gchar* hostname,
                guint port,
                C4sTcpResolvFunc resolv_func,
                C4sTcpConnectFunc connect_func,
                C4sTcpErrorFunc error_func,
                gpointer user_data,
                GError** error)
{
  C4sTcpConnector* connector;
  C4sTcpConnectorData* data;

  if(g_thread_supported() == FALSE)
    g_thread_init(NULL);

  connector = g_new(C4sTcpConnector, 1);
  data = g_new(C4sTcpConnectorData, 1);

  data->connector = connector;
  data->status = C4S_TCP_CONNECTOR_RUNNING;
  data->hostname = g_strdup(hostname);
  data->port = port;

  data->error = NULL;
  data->channel = NULL;

  connector->data = data;
  connector->resolv_func = resolv_func;
  connector->connect_func = connect_func;
  connector->error_func = error_func;
  connector->user_data = user_data;

  data->worker = g_thread_create(
    c4s_tcp_connector_thread_func,
    data,
    TRUE,
    error
  );

  if(data->worker == NULL)
  {
    g_free(data->hostname);
    g_free(data);
    g_free(connector);
    return NULL;
  }

  return connector;
}

void
c4s_tcp_cancel(C4sTcpConnector* connector)
{
  /* Do not free data since the thread is still running. The thread is
   * responsible for cleaning it up */
  connector->data->status = C4S_TCP_CONNECTOR_EXITED;
  g_free(connector);
}

GType
c4s_tcp_connection_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sTcpConnectionClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_tcp_connection_class_init,
      NULL,
      NULL,
      sizeof(C4sTcpConnection),
      0,
      (GInstanceInitFunc)c4s_tcp_connection_init
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sTcpConnection",
      &our_info,
      0
    );
  }

  return our_type;
}

GIOStatus
c4s_tcp_connection_send(C4sTcpConnection* conn,
                        const void* buf,
                        gsize size,
                        gsize* bytes_written,
                        GError** error)
{
  C4sTcpConnectionPrivate* priv;
  GIOStatus status;

  priv = C4S_TCP_CONNECTION_PRIVATE(conn);
  
  status = g_io_channel_write_chars(
    priv->channel,
    buf,
    size,
    bytes_written,
    error
  );

  g_io_channel_flush(priv->channel, error);

  return status;
}

GIOStatus
c4s_tcp_connection_recv(C4sTcpConnection* conn,
                        void* buf,
                        gsize size,
                        gsize* bytes_read,
                        GError** error)
{
  C4sTcpConnectionPrivate* priv;
  GIOStatus status;

  priv = C4S_TCP_CONNECTION_PRIVATE(conn);

  status = g_io_channel_read_chars(
    priv->channel,
    buf,
    size,
    bytes_read,
    error
  );

  return status;
}

void
c4s_tcp_connection_watch(C4sTcpConnection* conn,
                         GIOCondition condition,
                         C4sTcpIOFunc io_func,
                         gpointer user_data)
{
  C4sTcpConnectionPrivate* priv;
  priv = C4S_TCP_CONNECTION_PRIVATE(conn);

  if(priv->io_handle != 0)
    g_source_remove(priv->io_handle);

  if(condition == 0)
    return;

  priv->io_func = io_func;
  priv->user_data = user_data;

  priv->io_handle = g_io_add_watch(
    priv->channel,
    condition,
    c4s_tcp_connection_io_func,
    conn
  );

#ifdef _WIN32
  /* This is a workaround for a bug in glib that does not watch for the
   * given conditions until the next main loop iteration. This iteration is
   * made by this idle handler that does nothing. We could probably also call
   * gtk_main_iteration. */
  g_idle_add(c4s_tcp_connection_noop_func, NULL);
#endif
}
