/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/comm/c4s-comm.h>
#include <windows.h>
#include <gdk/gdkwin32.h>

struct _C4sCommBackend {
  C4sCommFileFunc file_func;
  gpointer user_data;

  HWND wnd;
};

static gboolean
c4s_comm_win32_open(C4sCommBackend* backend,
                    GtkWindow* window);

static GdkFilterReturn
c4s_comm_win32_filter(GdkXEvent* xevent,
                      GdkEvent* event,
                      gpointer user_data)
{
  C4sCommBackend* backend;
  MSG* msg;
  const COPYDATASTRUCT* cds;
  gchar* file;

  backend = user_data;
  msg = (MSG*)xevent;

  if(msg->message == WM_COPYDATA)
  {
    cds = (COPYDATASTRUCT*)msg->lParam;
    file = g_strndup((const gchar*)cds->lpData, cds->cbData);
    backend->file_func(file, backend->user_data);
    g_free(file);
    return GDK_FILTER_REMOVE;
  }
  
  return GDK_FILTER_CONTINUE;
}

static void
c4s_comm_win32_realized(GtkWindow* window,
                        gpointer user_data)
{
  C4sCommBackend* backend;
  backend = user_data;
  
  c4s_comm_win32_open(backend, window);
}

static C4sCommBackend*
c4s_comm_win32_init(C4sCommFileFunc file_func,
                    gpointer user_data)
{
  C4sCommBackend* backend;
  backend = g_new(C4sCommBackend, 1);
  
  backend->file_func = file_func;
  backend->user_data = user_data;
  
  backend->wnd = NULL;
  return backend;
}

static void
c4s_comm_win32_deinit(C4sCommBackend* backend)
{
  g_free(backend);
}

static gboolean
c4s_comm_win32_open(C4sCommBackend* backend,
                    GtkWindow* window)
{
  GdkWindow* gdk_window;
  HKEY key;
  gchar buf[32];
  int result;

  gdk_window = GTK_WIDGET(window)->window;

  if(gdk_window == NULL)
  {
    g_signal_connect_after(
      G_OBJECT(window),
      "realize",
      G_CALLBACK(c4s_comm_win32_realized),
      backend
    );

    return TRUE;
  }
  else
  {
    result = RegCreateKeyEx(
      HKEY_CURRENT_USER,
      "Software\\C4Scripter",
      0,
      NULL,
      REG_OPTION_NON_VOLATILE,
      KEY_ALL_ACCESS,
      NULL,
      &key,
      NULL
    );

   
    if(result != ERROR_SUCCESS)
      return FALSE;

    sprintf(buf, "%p", (HWND)GDK_WINDOW_HWND(gdk_window));
    result = RegSetValueEx(
      key,
      "Instance",
      0,
      REG_SZ,
      (BYTE*)buf,
      strlen(buf)+1
    );

    RegCloseKey(key);

    if(result != ERROR_SUCCESS)
      return FALSE;

    gdk_window_add_filter(gdk_window, c4s_comm_win32_filter, backend);
    return TRUE;
  }
}

static gboolean
c4s_comm_win32_connect(C4sCommBackend* backend)
{
  HKEY key;
  gchar buffer[32];
  DWORD len;
  DWORD type;
  int result;
  
  result = RegOpenKeyEx(
    HKEY_CURRENT_USER,
    "Software\\C4Scripter",
    0,
    KEY_ALL_ACCESS,
    &key
  );
  
  if(result != ERROR_SUCCESS)
    return FALSE;

  len = 32;
  result = RegQueryValueEx(
    key,
    "Instance",
    NULL,
    &type,
    (BYTE*)buffer,
    &len
  );

  RegCloseKey(key);

  if(result != ERROR_SUCCESS)
    return FALSE;

  if(type != REG_SZ) return FALSE;

  sscanf(buffer, "%p", &backend->wnd);

  if(!IsWindow(backend->wnd))
  {
    backend->wnd = NULL;
    return FALSE;
  }

  return TRUE;
}

static void
c4s_comm_win32_send(C4sCommBackend* backend,
                    int argc,
                    gchar* const * argv)
{
  int i;
  COPYDATASTRUCT cds;
  gchar* file;

  for(i = 0; i < argc; ++ i)
  {
    cds.dwData = 0;
    cds.cbData = strlen(argv[i]);

    file = g_malloc(cds.cbData * sizeof(gchar));
    memcpy(file, argv[i], cds.cbData);
    cds.lpData = file;

    SendMessage(backend->wnd, WM_COPYDATA, 0, (LPARAM)&cds);
    g_free(file);
  }
}

const C4sCommBackendFuncs C4S_COMM_BACKEND_WIN32 = {
  c4s_comm_win32_init,
  c4s_comm_win32_deinit,
  c4s_comm_win32_open,
  c4s_comm_win32_connect,
  c4s_comm_win32_send
};
