/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-enginefunctionlist.h>
#include <c4scripter/c4s-preferences.h>
#include <c4scripter/c4s-http.h>
#include <c4scripter/c4s-tcp.h>

#include <string.h>

struct _C4sLoaderBackend {
  C4sTcpConnector* connector;
  C4sTcpConnection* connection;
  C4sHttpRequest* http;

  GPtrArray* function_paths;
  
  const C4sLoaderBackendCallbacks* callbacks;
  gpointer user_data;
};

static const gsize ONLINE_FUNC_PREFIX = 14; /* strlen("sdk/script/fn/") */
static const gsize ABSOLUTE_FUNC_PREFIX = 23; /* strlen("/docs/de/sdk/...") */

static gchar*
c4s_loader_backend_online_get_function_name(C4sHttpRequest* http)
{
  const gchar* file;
  const gchar* sep;

  /* 9 is strlen("/docs/de/") */
  file = c4s_http_request_get_file(http);
  sep = strrchr(file + ABSOLUTE_FUNC_PREFIX, '.');
  g_assert(sep != NULL);

  return g_strndup(
    file + ABSOLUTE_FUNC_PREFIX,
    sep - file - ABSOLUTE_FUNC_PREFIX
  );
}

static void
c4s_loader_backend_online_function_done_cb(C4sHttpRequest* http,
                                           const gchar* text,
                                           gsize len,
                                           gpointer user_data)
{
  C4sLoaderBackend* backend;
  C4sEngineFunction* function;
  GError* error;
  gchar* name;

  backend = user_data;
  function = c4s_engine_function_new(text, len);  
  error = NULL;

  if(function == NULL)
  {
    g_set_error(
      &error,
      g_quark_from_static_string("C4S_LOADER_BACKEND_ONLINE"),
      C4S_LOADER_PARSE_ERROR,
      "Failed to parse %s",
      c4s_http_request_get_file(http)
    );
    
    name = c4s_loader_backend_online_get_function_name(http);

    g_object_unref(G_OBJECT(backend->http));
    backend->http = NULL;
    
    backend->callbacks->function_failed(name, error, backend->user_data);
    
    g_free(name);
    g_error_free(error);
  }
  else
  {
    g_object_unref(G_OBJECT(backend->http));
    backend->http = NULL;

    backend->callbacks->function_done(function, backend->user_data);
  }
}

static void
c4s_loader_backend_online_function_error_cb(C4sHttpRequest* http,
                                            const gchar* text,
                                            gpointer user_data)
{
  C4sLoaderBackend* backend;
  gchar* name;
  GError* error;

  backend = user_data;

  g_set_error(
    &error,
    g_quark_from_static_string("C4S_LOADER_BACKEND_ONLINE"),
    C4S_LOADER_TRANSPORT_ERROR,
    "Failure during function page download: %s",
    text
  );

  name = c4s_loader_backend_online_get_function_name(http);
  
  g_object_unref(G_OBJECT(backend->http));
  backend->http = NULL;
  
  backend->callbacks->function_failed(name, error, backend->user_data);

  g_free(name);
  g_error_free(error);
}

static void
c4s_loader_backend_online_index_done_cb(C4sHttpRequest* http,
                                        const gchar* text,
                                        gsize length,
                                        gpointer user_data)
{
  C4sLoaderBackend* backend;

  gboolean duplicate;
  const gchar* occ;
  const gchar* end;
  const gchar* sep;
  gsize pos;

  gsize len;
  gsize i;

  backend = user_data;
  pos = 0;

  while( (occ = g_strstr_len(text + pos, length - pos, "<a href=\"")) != NULL)
  {
    /* Point to beginning of link */
    occ += 9;

    end = memchr(occ, '\"', length - (occ - text));
    if(end == NULL) break;

    pos = end - text + 1;

    /* no script function */
    if(strncmp(occ, "sdk/script/fn/", ONLINE_FUNC_PREFIX) != 0)
      continue;

    /* Skip while, return, if and for because these are keywords rather
     * than script functions. */
    if(strncmp(occ + ONLINE_FUNC_PREFIX, "return.", 7) == 0)
      continue;
    if(strncmp(occ + ONLINE_FUNC_PREFIX, "while.", 6) == 0)
      continue;
    if(strncmp(occ + ONLINE_FUNC_PREFIX, "for.", 4) == 0)
      continue;
    if(strncmp(occ + ONLINE_FUNC_PREFIX, "if.", 3) == 0)
      continue;

    /* Strip trailing #Foo anchor if present */
    sep = memchr(
      occ + ONLINE_FUNC_PREFIX,
      '#',
      end - occ - ONLINE_FUNC_PREFIX
    );

    if(sep != NULL) end = sep;

    /* Check for duplicates */
    duplicate = FALSE;
    for(i = 0; i < backend->function_paths->len; ++ i)
    {
      len = strlen(backend->function_paths->pdata[i]);
      if(len == (gsize)(end - occ))
      {
        if(strncmp(backend->function_paths->pdata[i], occ, len) == 0)
        {
          duplicate = TRUE;
          break;
        }
      }
    }

    if(duplicate == FALSE)
      g_ptr_array_add(backend->function_paths, g_strndup(occ, end - occ));
  }

  g_object_unref(G_OBJECT(backend->http));
  backend->http = NULL;

  backend->callbacks->init_done(
    backend->function_paths->len,
    backend->user_data
  );
}

static void
c4s_loader_backend_online_index_error_cb(C4sHttpRequest* http,
                                         const gchar* text,
                                         gpointer user_data)
{
  C4sLoaderBackend* backend;
  GError* error;

  backend = user_data;

  g_object_unref(G_OBJECT(backend->http));
  backend->http = NULL;
  
  g_set_error(
    &error,
    g_quark_from_static_string("C4S_LOADER_BACKEND_ONLINE"),
    C4S_LOADER_TRANSPORT_ERROR,
    "Could not connect to clonk.de: %s",
    text
  );
  
  backend->callbacks->init_failed(error, backend->user_data);
  g_error_free(error);
}

static void
c4s_loader_backend_online_resolved_cb(gpointer user_data)
{
  C4sLoaderBackend* backend;
  backend = user_data;
  
  backend->callbacks->set_status(
    "Connecting to clonk.de",
    backend->user_data
  );
}

static void
c4s_loader_backend_online_connected_cb(C4sTcpConnection* conn,
                                       gpointer user_data)
{
  C4sLoaderBackend* backend;
  backend = user_data;
  
  backend->connector = NULL;
  backend->connection = conn;
  g_object_ref(G_OBJECT(backend->connection));
  
  backend->callbacks->set_status(
    "Requesting function index",
    backend->user_data
  );
  
  backend->http = c4s_http_request_new(
    conn,
    "/docs/de/content.html",
    c4s_loader_backend_online_index_done_cb,
    c4s_loader_backend_online_index_error_cb,
    backend
  );
}

static void
c4s_loader_backend_online_error_cb(const GError* error,
                                   gpointer user_data)
{
  C4sLoaderBackend* backend;
  backend = user_data;

  backend->callbacks->init_failed(error, backend->user_data);
}

static gboolean
c4s_loader_backend_online_init_idle(gpointer user_data)
{
  C4sLoaderBackend* backend;
  GError* error;

  backend = user_data;
  error = NULL;
  
  if(c4s_tcp_init(&error) == FALSE)
  {
    backend->callbacks->init_failed(error, backend->user_data);
  }
  else
  {
    backend->callbacks->set_status("Resolving clonk.de", backend->user_data);
    backend->connector = c4s_tcp_connect(
      "clonk.de",
      80,
      c4s_loader_backend_online_resolved_cb,
      c4s_loader_backend_online_connected_cb,
      c4s_loader_backend_online_error_cb,
      backend,
      &error
    );
    
    if(backend->connector == NULL)
    {
      backend->callbacks->init_failed(error, backend->user_data);
    }
  }

  return FALSE;
}

static C4sLoaderBackend*
c4s_loader_backend_online_init(C4sPreferences* preferences,
                               const C4sLoaderBackendCallbacks* callbacks,
                               gpointer user_data)
{
  C4sLoaderBackend* backend;
  backend = g_new(C4sLoaderBackend, 1);
  
  backend->connector = NULL;
  
  backend->callbacks = callbacks;
  backend->user_data = user_data;
  
  backend->function_paths = g_ptr_array_new();

  g_idle_add(c4s_loader_backend_online_init_idle, backend);
	return backend;
}

static void
c4s_loader_backend_online_deinit(C4sLoaderBackend* backend)
{
  gsize i;

  if(backend->connector != NULL)
    c4s_tcp_cancel(backend->connector);
  if(backend->connection != NULL)
    g_object_unref(G_OBJECT(backend->connection));
  if(backend->http != NULL)
    g_object_unref(G_OBJECT(backend->http));

  for(i = 0; i < backend->function_paths->len; ++ i)
    g_free(backend->function_paths->pdata[i]);
  g_ptr_array_free(backend->function_paths, TRUE);

  g_free(backend);
}

static void
c4s_loader_backend_online_load_function(C4sLoaderBackend* backend,
                                        unsigned int i)
{
  char* url;

  g_assert(backend->connection != NULL);
  g_assert(backend->http == NULL);
  
  url = g_strconcat("/docs/de/", backend->function_paths->pdata[i], NULL);

  backend->http = c4s_http_request_new(
    backend->connection,
    url,
    c4s_loader_backend_online_function_done_cb,
    c4s_loader_backend_online_function_error_cb,
    backend
  );
  
  g_free(url);
}

const C4sLoaderBackendFuncs C4S_LOADER_BACKEND_ONLINE = {
  c4s_loader_backend_online_init,
  c4s_loader_backend_online_deinit,
  c4s_loader_backend_online_load_function
};
