/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-preferences.h>

#include <glib/gfileutils.h>
#include <glib/gkeyfile.h>
#include <glib/gutils.h>

#include <errno.h>
#include <string.h>

struct _C4sStorageBackend {
	gchar* file_path;
	GKeyFile* key_file;
};

static C4sStorageBackend*
c4s_storage_keyfile_init(GError** error)
{
  C4sStorageBackend* backend;
  backend = g_new(C4sStorageBackend, 1);

  backend->file_path = g_build_filename(
    g_get_home_dir(),
    ".c4scripter",
    "config",
    NULL
  );

  backend->key_file = g_key_file_new();

  /* Note that the file not existing is not an error since it might be
   * the first time the program is launched */
  g_key_file_load_from_file(backend->key_file, backend->file_path, 0, NULL);
  return backend;
}

static void
c4s_storage_keyfile_deinit(C4sStorageBackend* backend)
{
  GError* error;
  gsize length;
  gchar* data;

  gchar* path;
  path = g_path_get_dirname(backend->file_path);

  if(g_mkdir_with_parents(path, 0755) == -1)
  {
    g_warning("Could not create directory %s: %s\n", path, strerror(errno));
  }
  else
  {
    error = NULL;
    data = g_key_file_to_data(backend->key_file, &length, &error);
    if(data == NULL)
    {
      g_warning("Could not write config file: %s\n", error->message);
      g_error_free(error);
    }
    else
    {
      if(g_file_set_contents(backend->file_path, data, length, &error) == FALSE)
      {
        g_warning("Could not write config file: %s\n", error->message);
        g_error_free(error);
      }

      g_free(data);
    }
  }

  g_free(path);
  g_free(backend->file_path);
  g_key_file_free(backend->key_file);
  g_free(backend);
}

static gint
c4s_storage_keyfile_read_int(C4sStorageBackend* backend,
                             const gchar* group,
                             const gchar* key,
                             gint default_value)
{
  GError* error;
  gint value;

  error = NULL;
  value = g_key_file_get_integer(backend->key_file, group, key, &error);
  if(error != NULL)
  {
    g_error_free(error);
    return default_value;
  }

  return value;
}

static gboolean
c4s_storage_keyfile_read_boolean(C4sStorageBackend* backend,
                                 const gchar* group,
                                 const gchar* key,
                                 gboolean default_value)
{
  GError* error;
  gboolean value;

  error = NULL;
  value = g_key_file_get_boolean(backend->key_file, group, key, &error);
  if(error != NULL)
  {
    g_error_free(error);
    return default_value;
  }

  return value;
}

static gchar*
c4s_storage_keyfile_read_string(C4sStorageBackend* backend,
                                const gchar* group,
                                const gchar* key,
                                const gchar* default_value)
{
  GError* error;
  gchar* value;

  error = NULL;
  value = g_key_file_get_string(backend->key_file, group, key, &error);
  if(error != NULL)
  {
    g_error_free(error);
    return g_strdup(default_value);
  }

  return value;
}

static void
c4s_storage_keyfile_write_int(C4sStorageBackend* backend,
                              const gchar* group,
                              const gchar* key,
                              gint value)
{
  g_key_file_set_integer(backend->key_file, group, key, value);
}

static void
c4s_storage_keyfile_write_boolean(C4sStorageBackend* backend,
                                  const gchar* group,
                                  const gchar* key,
                                  gboolean value)
{
  g_key_file_set_boolean(backend->key_file, group, key, value);
}

static void
c4s_storage_keyfile_write_string(C4sStorageBackend* backend,
                                 const gchar* group,
                                 const gchar* key,
                                 const gchar* value)
{
  g_key_file_set_string(backend->key_file, group, key, value);
}

const C4sStorageBackendFuncs C4S_STORAGE_KEYFILE = {
	c4s_storage_keyfile_init,
	c4s_storage_keyfile_deinit,
	c4s_storage_keyfile_read_int,
	c4s_storage_keyfile_read_boolean,
	c4s_storage_keyfile_read_string,
	c4s_storage_keyfile_write_int,
	c4s_storage_keyfile_write_boolean,
	c4s_storage_keyfile_write_string
};
